"""
Overview:
    A model for rating anime images into 4 classes (``general``, ``sensitive``, ``questionable`` and ``explicit``),
    based on danbooru rating system.

    The following are sample images for testing.

    .. collapse:: The following are sample images for testing. (WARNING: NSFW!!!)

        .. image:: dbrating.plot.py.svg
            :align: center

    This is an overall benchmark of all the rating validation models:

    .. image:: dbrating_benchmark.plot.py.svg
        :align: center

    The models are hosted on
    `huggingface - deepghs/anime_dbrating <https://huggingface.co/deepghs/anime_dbrating>`_.

    .. note::
        This model is based on danbooru rating system, trained with 1.2 million images.
        If you need 3-level rating prediction, use :func:`imgutils.validate.rating.anime_rating`.

    .. note::
        Please note that the classification of ``general``, ``sensitive``, ``questionable`` and ``explicit`` types
        does not have clear boundaries, making it challenging to clean the training data. As a result,
        there is no strict ground truth for the rating classification problem. The judgment functionality
        provided by the current module is intended as a quick and rough estimation.

        **If you require an accurate filtering or judgment function specifically for R-18 images,
        it is recommended to consider using object detection-based methods**,
        such as using :func:`imgutils.detect.censor.detect_censors` to detect sensitive regions as the basis for judgment.
"""
from typing import Tuple, Dict

from ..data import ImageTyping
from ..generic import classify_predict, classify_predict_score

__all__ = [
    'anime_dbrating_score',
    'anime_dbrating',
]

_DEFAULT_MODEL_NAME = 'mobilenetv3_large_100_v0_ls0.2'
_REPO_ID = 'deepghs/anime_dbrating'


def anime_dbrating_score(image: ImageTyping, model_name: str = _DEFAULT_MODEL_NAME) -> Dict[str, float]:
    """
    Overview:
        Predict the rating of the given image, return the score with as a dict object.

    :param image: Image to rating.
    :param model_name: Model to use. Default is ``mobilenetv3_large_100_v0_ls0.2``.
        All available models are listed on the benchmark plot above.
        If you need better accuracy, just set this to ``caformer_s36_v0_ls0.2``.
    :return: A dict with ratings and scores.

    Examples::
        >>> from imgutils.validate import anime_dbrating
        >>>
        >>> anime_dbrating('general/1.jpg')
        ('general', 0.7508869767189026)
        >>> anime_dbrating('general/2.jpg')
        ('general', 0.7034655809402466)
        >>> anime_dbrating('general/3.jpg')
        ('general', 0.728887677192688)
        >>> anime_dbrating('general/4.jpg')
        ('general', 0.7404400110244751)
        >>> anime_dbrating('sensitive/5.jpg')
        ('sensitive', 0.7446154952049255)
        >>> anime_dbrating('sensitive/6.jpg')
        ('sensitive', 0.7514738440513611)
        >>> anime_dbrating('sensitive/7.jpg')
        ('sensitive', 0.768704354763031)
        >>> anime_dbrating('sensitive/8.jpg')
        ('sensitive', 0.8219676613807678)
        >>> anime_dbrating('questionable/9.jpg')
        ('questionable', 0.7267540693283081)
        >>> anime_dbrating('questionable/10.jpg')
        ('questionable', 0.7645740509033203)
        >>> anime_dbrating('questionable/11.jpg')
        ('questionable', 0.7216582894325256)
        >>> anime_dbrating('questionable/12.jpg')
        ('questionable', 0.7615436315536499)
        >>> anime_dbrating('explicit/13.jpg')
        ('explicit', 0.815083920955658)
        >>> anime_dbrating('explicit/14.jpg')
        ('explicit', 0.8321858644485474)
        >>> anime_dbrating('explicit/15.jpg')
        ('explicit', 0.8204999566078186)
        >>> anime_dbrating('explicit/16.jpg')
        ('explicit', 0.820833146572113)

    """
    return classify_predict_score(image, _REPO_ID, model_name)


def anime_dbrating(image: ImageTyping, model_name: str = _DEFAULT_MODEL_NAME) -> Tuple[str, float]:
    """
    Overview:
        Predict the rating of the given image, return the class and its score.

    :param image: Image to rating.
    :param model_name: Model to use. Default is ``mobilenetv3_large_100_v0_ls0.2``. All available models are listed
        on the benchmark plot above. If you need better accuracy, just set this to ``caformer_s36_v0_ls0.2``.
    :return: A tuple contains the rating and its score.

    Examples::
        >>> from imgutils.validate import anime_dbrating_score
        >>>
        >>> os.chdir('docs/source/api_doc/validate/dbrating')
        >>>
        >>> anime_dbrating_score('general/1.jpg')
        {'general': 0.7508870363235474, 'sensitive': 0.11212056130170822, 'questionable': 0.06781744956970215, 'explicit': 0.06917501986026764}
        >>> anime_dbrating_score('general/2.jpg')
        {'general': 0.7034654021263123, 'sensitive': 0.15903906524181366, 'questionable': 0.06688199192285538, 'explicit': 0.07061357796192169}
        >>> anime_dbrating_score('general/3.jpg')
        {'general': 0.7288877964019775, 'sensitive': 0.1476859599351883, 'questionable': 0.060362350195646286, 'explicit': 0.06306383013725281}
        >>> anime_dbrating_score('general/4.jpg')
        {'general': 0.7404399514198303, 'sensitive': 0.10337048768997192, 'questionable': 0.08087948709726334, 'explicit': 0.07530999928712845}
        >>> anime_dbrating_score('sensitive/5.jpg')
        {'general': 0.055992450565099716, 'sensitive': 0.7446154356002808, 'questionable': 0.13191790878772736, 'explicit': 0.06747424602508545}
        >>> anime_dbrating_score('sensitive/6.jpg')
        {'general': 0.06458679586648941, 'sensitive': 0.7514738440513611, 'questionable': 0.10566363483667374, 'explicit': 0.07827574014663696}
        >>> anime_dbrating_score('sensitive/7.jpg')
        {'general': 0.07079866528511047, 'sensitive': 0.7687042951583862, 'questionable': 0.09974884241819382, 'explicit': 0.06074819341301918}
        >>> anime_dbrating_score('sensitive/8.jpg')
        {'general': 0.050435908138751984, 'sensitive': 0.8219675421714783, 'questionable': 0.0593985915184021, 'explicit': 0.06819795072078705}
        >>> anime_dbrating_score('questionable/9.jpg')
        {'general': 0.06569571048021317, 'sensitive': 0.1177448257803917, 'questionable': 0.726753830909729, 'explicit': 0.08980562537908554}
        >>> anime_dbrating_score('questionable/10.jpg')
        {'general': 0.06481882929801941, 'sensitive': 0.06922297924757004, 'questionable': 0.7645740509033203, 'explicit': 0.10138414055109024}
        >>> anime_dbrating_score('questionable/11.jpg')
        {'general': 0.06351721286773682, 'sensitive': 0.07683827728033066, 'questionable': 0.7216582894325256, 'explicit': 0.13798624277114868}
        >>> anime_dbrating_score('questionable/12.jpg')
        {'general': 0.05942752957344055, 'sensitive': 0.10584963858127594, 'questionable': 0.7615437507629395, 'explicit': 0.07317910343408585}
        >>> anime_dbrating_score('explicit/13.jpg')
        {'general': 0.060196295380592346, 'sensitive': 0.06751583516597748, 'questionable': 0.0572039857506752, 'explicit': 0.815083920955658}
        >>> anime_dbrating_score('explicit/14.jpg')
        {'general': 0.05398125201463699, 'sensitive': 0.06124086305499077, 'questionable': 0.0525919646024704, 'explicit': 0.8321859240531921}
        >>> anime_dbrating_score('explicit/15.jpg')
        {'general': 0.05922013148665428, 'sensitive': 0.06274889409542084, 'questionable': 0.057530902326107025, 'explicit': 0.8205001354217529}
        >>> anime_dbrating_score('explicit/16.jpg')
        {'general': 0.05683052912354469, 'sensitive': 0.06635929644107819, 'questionable': 0.05597696080803871, 'explicit': 0.8208332657814026}
    """
    return classify_predict(image, _REPO_ID, model_name)



"""
Overview:
    Upscale images with CDC model, developed and trained by `7eu7d7 <https://github.com/7eu7d7>`_,
    the models are hosted on `deepghs/cdc_anime_onnx <https://huggingface.co/deepghs/cdc_anime_onnx>`_.

    Here are some examples:

    .. image:: cdc_demo.plot.py.svg
        :align: center

    Here is the benchmark of CDC models:

    .. image:: cdc_benchmark.plot.py.svg
        :align: center

    .. note::
        CDC model has high quality, and really low running speed.
        As we tested, when it upscales an image with 1024x1024 resolution on 2060 GPU,
        the time cost is approx 70s/image. So we strongly recommend against running it on CPU.
        Please run CDC model on environments with GPU for better experience.
"""
from typing import Tuple, Any

import numpy as np
from PIL import Image
from huggingface_hub import hf_hub_download

from ..data import ImageTyping
from ..generic import ImageEnhancer
from ..utils import open_onnx_model, area_batch_run, ts_lru_cache

__all__ = [
    'upscale_with_cdc',
]


@ts_lru_cache()
def _open_cdc_upscaler_model(model: str) -> Tuple[Any, int]:
    """
    Opens and initializes the CDC upscaler model.

    :param model: The name of the model to use.
    :type model: str

    :return: Tuple of the ONNX model and the scale factor.
    :rtype: Tuple[Any, int]
    """
    ort = open_onnx_model(hf_hub_download(
        f'deepghs/cdc_anime_onnx',
        f'{model}.onnx'
    ))

    input_ = np.random.randn(1, 3, 16, 16).astype(np.float32)
    output_, = ort.run(['output'], {'input': input_})

    batch, channels, scale_h, height, scale_w, width = output_.shape
    assert batch == 1 and channels == 3 and height == 16 and width == 16, \
        f'Unexpected output size found {output_.shape!r}.'
    assert scale_h == scale_w, f'Scale of height and width not match - {output_.shape!r}.'

    return ort, scale_h


_CDC_INPUT_UNIT = 16


def _upscale_for_rgb(input_: np.ndarray, model: str = 'HGSR-MHR-anime-aug_X4_320',
                     tile_size: int = 512, tile_overlap: int = 64, batch_size: int = 1, silent: bool = False):
    assert len(input_.shape) == 4 and input_.shape[:2] == (1, 3)
    ort, scale = _open_cdc_upscaler_model(model)

    def _method(ix):
        ix = ix.astype(np.float32)
        batch, channels, height, width = ix.shape
        p_height = 0 if height % _CDC_INPUT_UNIT == 0 else _CDC_INPUT_UNIT - (height % _CDC_INPUT_UNIT)
        p_width = 0 if width % _CDC_INPUT_UNIT == 0 else _CDC_INPUT_UNIT - (width % _CDC_INPUT_UNIT)
        if p_height > 0 or p_width > 0:  # align to 16
            ix = np.pad(ix, ((0, 0), (0, 0), (0, p_height), (0, p_width)), mode='reflect')
        actual_height, actual_width = height, width

        ox, = ort.run(['output'], {'input': ix})
        batch, channels, scale_, height, scale_, width = ox.shape
        ox = ox.reshape((batch, channels, scale_ * height, scale_ * width))
        ox = ox[..., :scale_ * actual_height, :scale_ * actual_width]  # crop back
        return ox

    output_ = area_batch_run(
        input_, _method,
        tile_size=tile_size, tile_overlap=tile_overlap, batch_size=batch_size,
        scale=scale, silent=silent, process_title='CDC Upscale',
    )
    output_ = np.clip(output_, a_min=0.0, a_max=1.0)
    return output_


class _Enhancer(ImageEnhancer):
    def __init__(self, model: str = 'HGSR-MHR-anime-aug_X4_320',
                 tile_size: int = 512, tile_overlap: int = 64, batch_size: int = 1, silent: bool = False):
        self.model = model
        self.tile_size = tile_size
        self.tile_overlap = tile_overlap
        self.batch_size = batch_size
        self.silent = silent

    def _process_rgb(self, rgb_array: np.ndarray):
        return _upscale_for_rgb(
            rgb_array[None, ...],
            model=self.model,
            tile_size=self.tile_size,
            tile_overlap=self.tile_overlap,
            batch_size=self.batch_size,
            silent=self.silent,
        )[0]


@ts_lru_cache()
def _get_enhancer(model: str = 'HGSR-MHR-anime-aug_X4_320',
                  tile_size: int = 512, tile_overlap: int = 64, batch_size: int = 1, silent: bool = False):
    return _Enhancer(model, tile_size, tile_overlap, batch_size, silent)


def upscale_with_cdc(image: ImageTyping, model: str = 'HGSR-MHR-anime-aug_X4_320',
                     tile_size: int = 512, tile_overlap: int = 64, batch_size: int = 1,
                     silent: bool = False) -> Image.Image:
    """
    Upscale the input image using the CDC upscaler model.

    :param image: The input image.
    :type image: ImageTyping

    :param model: The name of the model to use. (default: 'HGSR-MHR-anime-aug_X4_320')
    :type model: str

    :param tile_size: The size of each tile. (default: 512)
    :type tile_size: int

    :param tile_overlap: The overlap between tiles. (default: 64)
    :type tile_overlap: int

    :param batch_size: The batch size. (default: 1)
    :type batch_size: int

    :param silent: Whether to suppress progress messages. (default: False)
    :type silent: bool

    :return: The upscaled image.
    :rtype: Image.Image

    .. note::
        RGBA images are supported. When you pass an image with transparency channel (e.g. RGBA image),
        this function will return an RGBA image, otherwise return a RGB image.

    Example::
        >>> from PIL import Image
        >>> from imgutils.upscale import upscale_with_cdc
        >>>
        >>> image = Image.open('cute_waifu_aroma.png')
        >>> image
        <PIL.PngImagePlugin.PngImageFile image mode=RGBA size=1168x1168 at 0x7F0E8CA06880>
        >>>
        >>> upscale_with_cdc(image)
        <PIL.Image.Image image mode=RGBA size=4672x4672 at 0x7F0E48EDB640>
    """
    return _get_enhancer(model, tile_size, tile_overlap, batch_size, silent).process(image)



"""
Exception class used to indicate that a transform object cannot be parsed by a specific parser.
"""


class NotParseTarget(Exception):
    """
    Exception raised when a transform object cannot be parsed by a specific parser.

    This exception is used internally by transform parsers to indicate that they are not
    able to handle the given transform object, allowing the parsing system to try the
    next available parser.
    """
    pass



"""
Overview:
    Tagging utils based on ML-danbooru which is provided by 7eu7d7. The code is here:
    `7eu7d7/ML-Danbooru <https://github.com/7eu7d7/ML-Danbooru>`_ .
"""
from typing import Tuple, List

import numpy as np
import pandas as pd
from PIL import Image
from huggingface_hub import hf_hub_download

from .overlap import drop_overlap_tags
from ..data import load_image, ImageTyping
from ..utils import open_onnx_model, ts_lru_cache


@ts_lru_cache()
def _open_mldanbooru_model():
    return open_onnx_model(hf_hub_download('deepghs/ml-danbooru-onnx', 'ml_caformer_m36_dec-5-97527.onnx'))


def _resize_align(image: Image.Image, size: int, keep_ratio: float = True, align: int = 4) -> Image.Image:
    if not keep_ratio:
        target_size = (size, size)
    else:
        min_edge = min(image.size)
        target_size = (
            int(image.size[0] / min_edge * size),
            int(image.size[1] / min_edge * size),
        )

    target_size = (
        (target_size[0] // align) * align,
        (target_size[1] // align) * align,
    )

    return image.resize(target_size, resample=Image.BILINEAR)


def _to_tensor(image: Image.Image):
    # noinspection PyTypeChecker
    img: np.ndarray = np.array(image, dtype=np.uint8, copy=True)
    img = img.reshape((image.size[1], image.size[0], len(image.getbands())))

    # put it from HWC to CHW format
    img = img.transpose((2, 0, 1))
    return img.astype(np.float32) / 255


@ts_lru_cache()
def _get_mldanbooru_labels(use_real_name: bool = False) -> Tuple[List[str], List[int], List[int]]:
    path = hf_hub_download('deepghs/imgutils-models', 'mldanbooru/mldanbooru_tags.csv')
    df = pd.read_csv(path)

    return df["name"].tolist() if not use_real_name else df['real_name'].tolist()


def get_mldanbooru_tags(image: ImageTyping, use_real_name: bool = False,
                        threshold: float = 0.7, size: int = 448, keep_ratio: bool = False,
                        drop_overlap: bool = False):
    """
    Overview:
        Tagging image with ML-Danbooru, similar to
        `deepghs/ml-danbooru-demo <https://huggingface.co/spaces/deepghs/ml-danbooru-demo>`_.

    :param image: Image to tagging.
    :param use_real_name: Use real name on danbooru. Due to the renaming and redirection of many tags
        on the Danbooru website after the training of ``deepdanbooru``,
        it may be necessary to use the latest tag names in some application scenarios.
        The default value of ``False`` indicates the use of the original tag names.
    :param threshold: Threshold for tags, default is ``0.7``.
    :param size: Size when passing the resized image into model, default is ``448``.
    :param keep_ratio: Keep the original ratio between height and width when passing the image into
        model, default is ``False``.
    :param drop_overlap: Drop overlap tags or not, default is ``False``.

    Example:
        Here are some images for example

        .. image:: tagging_demo.plot.py.svg
           :align: center

        >>> import os
        >>> from imgutils.tagging import get_mldanbooru_tags
        >>>
        >>> get_mldanbooru_tags('skadi.jpg')
        {'1girl': 0.9999984502792358, 'long_hair': 0.9999946355819702, 'red_eyes': 0.9994951486587524, 'navel': 0.998144268989563, 'breasts': 0.9978417158126831, 'solo': 0.9941409230232239, 'shorts': 0.9799384474754333, 'gloves': 0.979142427444458, 'very_long_hair': 0.961823582649231, 'looking_at_viewer': 0.961323618888855, 'silver_hair': 0.9490893483161926, 'large_breasts': 0.9450850486755371, 'midriff': 0.9425153136253357, 'sweat': 0.9409335255622864, 'thighs': 0.9319437146186829, 'crop_top': 0.9265308976173401, 'baseball_bat': 0.9259042143821716, 'sky': 0.922250509262085, 'holding': 0.9199565052986145, 'outdoors': 0.9175475835800171, 'day': 0.9102761745452881, 'black_gloves': 0.9076938629150391, 'stomach': 0.9052775502204895, 'shirt': 0.8938589692115784, 'cowboy_shot': 0.8894285559654236, 'bangs': 0.8891903162002563, 'blue_sky': 0.8845980763435364, 'parted_lips': 0.8842408061027527, 'hair_between_eyes': 0.8659475445747375, 'sportswear': 0.862621009349823, 'no_headwear': 0.8616052865982056, 'cloud': 0.8562789559364319, 'short_shorts': 0.8555729389190674, 'no_hat': 0.8533340096473694, 'black_shorts': 0.8477485775947571, 'short_sleeves': 0.8430152535438538, 'low-tied_long_hair': 0.8340626955032349, 'crop_top_overhang': 0.8266023397445679, 'holding_baseball_bat': 0.8222048282623291, 'standing': 0.8202669620513916, 'black_shirt': 0.8061150312423706, 'ass_visible_through_thighs': 0.7803354859352112, 'thigh_gap': 0.7789446711540222, 'arms_up': 0.7052110433578491}
        >>>
        >>> get_mldanbooru_tags('hutao.jpg')
        {'1girl': 0.9999866485595703, 'skirt': 0.997043788433075, 'tongue': 0.9969649910926819, 'hair_ornament': 0.9957101345062256, 'tongue_out': 0.9928386807441711, 'flower': 0.9886980056762695, 'twintails': 0.9864778518676758, 'ghost': 0.9769423007965088, 'hair_flower': 0.9747489094734192, 'bag': 0.9736957550048828, 'long_hair': 0.9388670325279236, 'backpack': 0.9356311559677124, 'brown_hair': 0.91000896692276, 'cardigan': 0.8955123424530029, 'red_eyes': 0.8910233378410339, 'plaid': 0.8904104828834534, 'looking_at_viewer': 0.8881211280822754, 'school_uniform': 0.8876776695251465, 'outdoors': 0.8864808678627014, 'jacket': 0.8810517191886902, 'plaid_skirt': 0.8798807263374329, 'ahoge': 0.8765745162963867, 'pleated_skirt': 0.8737136125564575, 'nail_polish': 0.8650439381599426, 'solo': 0.8613706827163696, 'blue_cardigan': 0.8571277260780334, 'bangs': 0.8333670496940613, 'very_long_hair': 0.8160212635993958, 'eyebrows_visible_through_hair': 0.8122442364692688, 'hairclip': 0.8091571927070618, 'red_nails': 0.8082079887390137, ':p': 0.8048468232154846, 'long_sleeves': 0.8042327165603638, 'shirt': 0.7984272241592407, 'blazer': 0.794708251953125, 'ribbon': 0.78981614112854, 'hair_ribbon': 0.7892146110534668, 'star-shaped_pupils': 0.7867060899734497, 'gradient_hair': 0.786359965801239, 'white_shirt': 0.7790888547897339, 'brown_skirt': 0.7760675549507141, 'symbol-shaped_pupils': 0.774523913860321, 'smile': 0.7721588015556335, 'hair_between_eyes': 0.7697228789329529, 'cowboy_shot': 0.755959689617157, 'multicolored_hair': 0.7477189898490906, 'blush': 0.7476690411567688, 'railing': 0.7476617693901062, 'blue_jacket': 0.7458406090736389, 'sleeves_past_wrists': 0.741143524646759, 'day': 0.7364678978919983, 'collared_shirt': 0.7193643450737, 'red_neckwear': 0.7108616828918457, 'flower-shaped_pupils': 0.7086325287818909, 'miniskirt': 0.7055293321609497, 'holding': 0.7039415836334229, 'open_clothes': 0.7018357515335083}

    .. note::
        ML-Danbooru only contains generic tags, so the return value will not be splitted like that in
        :func:`imgutils.tagging.deepdanbooru.get_deepdanbooru_tags` or
        :func:`imgutils.tagging.wd14.get_wd14_tags`.
    """
    image = load_image(image, mode='RGB')
    real_input = _to_tensor(_resize_align(image, size, keep_ratio))
    real_input = real_input.reshape(1, *real_input.shape)

    model = _open_mldanbooru_model()
    native_output, = model.run(['output'], {'input': real_input})

    output = (1 / (1 + np.exp(-native_output))).reshape(-1)
    tags = _get_mldanbooru_labels(use_real_name)
    pairs = sorted([(tags[i], ratio) for i, ratio in enumerate(output)], key=lambda x: (-x[1], x[0]))

    general_tags = {tag: float(ratio) for tag, ratio in pairs if ratio >= threshold}
    if drop_overlap:
        general_tags = drop_overlap_tags(general_tags)
    return general_tags



"""
Overview:
    Meta information for imgutils package.
"""

#: Title of this project (should be `imgutils`).
__TITLE__ = 'imgutils'

#: Version of this project.
__VERSION__ = '0.9.1'

#: Short description of the project, will be included in ``setup.py``.
__DESCRIPTION__ = 'A convenient and user-friendly anime-style image data processing library that integrates ' \
                  'various advanced anime-style image processing models.'

#: Author of this project.
__AUTHOR__ = 'narugo1992, 7eu7d7'

#: Email of the authors'.
__AUTHOR_EMAIL__ = 'narugo1992@deepghs.org'



"""
Overview:
    Get edge with lineart anime model.
"""
from functools import partial
from typing import Optional

import numpy as np
from huggingface_hub import hf_hub_download

from ._base import resize_image, cv2_resize, _get_image_edge
from ..data import ImageTyping, load_image
from ..utils import open_onnx_model, ts_lru_cache


def _preprocess(input_image, detect_resolution: int = 512):
    input_image = np.array(input_image, dtype=np.uint8)
    img = resize_image(input_image, detect_resolution, align=256).astype(np.float32)
    img = (img / 127.5 - 1.0).transpose(2, 0, 1)[None, ...]
    return img


@ts_lru_cache()
def _open_la_anime_model():
    return open_onnx_model(hf_hub_download(
        'deepghs/imgutils-models',
        f'lineart/lineart_anime.onnx',
    ))


def get_edge_by_lineart_anime(image: ImageTyping, detect_resolution: int = 512):
    """
    Overview:
        Get edge mask with lineart anime model.

    :param image: Original image (assuming its size is ``HxW``).
    :param detect_resolution: Resolution when passing the image into neural network. Default is ``512``.
    :return: A mask with format ``float32[H, W]``.
    """
    image = load_image(image, mode='RGB')
    output_, = _open_la_anime_model().run(['output'], {'input': _preprocess(image, detect_resolution)})
    output_ = (output_ + 1.0) / 2.0
    output_ = cv2_resize(output_[0].transpose(1, 2, 0), image.width, image.height)
    return 1.0 - output_.clip(0.0, 1.0)


def edge_image_with_lineart_anime(image: ImageTyping, detect_resolution: int = 512,
                                  backcolor: str = 'white', forecolor: Optional[str] = None):
    """
    Overview:
        Get an image with the extracted edge from ``image``.

    :param image: Original image (assuming its size is ``HxW``).
    :param detect_resolution: Resolution when passing the image into neural network. Default is ``512``.
    :param backcolor: Background color the target image. Default is ``white``. When ``transparent`` is given, \
        the background will be transparent.
    :param forecolor: Fore color of the target image. Default is ``None`` which means use the color \
        from the given ``image``.
    :return: An image with the extracted edge from ``image``.

    Examples::
        .. image:: lineart_anime.plot.py.svg
            :align: center
    """
    return _get_image_edge(
        image,
        partial(get_edge_by_lineart_anime, detect_resolution=detect_resolution),
        backcolor, forecolor
    )



"""
This module provides functionality for YOLO object detection using ONNX models from Hugging Face.

It includes utilities for preprocessing images, performing object detection, and post-processing
the results. The main components are:

1. YOLOModel class: Manages YOLO models from a Hugging Face repository.
2. Helper functions for coordinate conversion, non-maximum suppression, and image processing.
3. A high-level function 'yolo_predict' for easy object detection on images.

The module supports various image input types and allows customization of confidence and IoU thresholds.
"""

import ast
import json
import math
import os
from threading import Lock
from typing import List, Optional, Tuple, Union

import numpy as np
from PIL import Image
from hbutils.color import rnd_colors
from hfutils.operate import get_hf_client, get_hf_fs
from hfutils.repository import hf_hub_repo_url
from hfutils.utils import hf_fs_path, hf_normpath
from huggingface_hub import HfFileSystem, hf_hub_download

from ..data import load_image, rgb_encode, ImageTyping
from ..utils import open_onnx_model, ts_lru_cache

try:
    import gradio as gr
except (ImportError, ModuleNotFoundError):
    gr = None

__all__ = [
    'YOLOModel',
    'yolo_predict',
]


def _check_gradio_env():
    """
    Check if the Gradio library is installed and available.

    :raises EnvironmentError: If Gradio is not installed.
    """
    if gr is None:
        raise EnvironmentError(f'Gradio required for launching webui-based demo.\n'
                               f'Please install it with `pip install dghs-imgutils[demo]`.')


def _v_fix(v):
    """
    Round and convert a float value to an integer.

    :param v: The float value to be rounded and converted.
    :type v: float
    :return: The rounded integer value.
    :rtype: int
    """
    return int(round(v))


def _bbox_fix(bbox):
    """
    Fix the bounding box coordinates by rounding them to integers.

    :param bbox: The bounding box coordinates.
    :type bbox: tuple
    :return: A tuple of fixed (rounded to integer) bounding box coordinates.
    :rtype: tuple
    """
    return tuple(map(_v_fix, bbox))


def _yolo_xywh2xyxy(x: np.ndarray) -> np.ndarray:
    """
    Convert bounding box coordinates from (x, y, width, height) format to (x1, y1, x2, y2) format.

    This function is adapted from YOLOv8 and transforms the center-based representation
    to a corner-based representation of bounding boxes.

    :param x: Input bounding box coordinates in (x, y, width, height) format.
    :type x: np.ndarray

    :return: Bounding box coordinates in (x1, y1, x2, y2) format.
    :rtype: np.ndarray

    :Example:

    >>> import numpy as np
    >>> boxes = np.array([[10, 10, 20, 20]])
    >>> _yolo_xywh2xyxy(boxes)
    array([[  0.,   0.,  20.,  20.]])
    """
    y = np.copy(x)
    y[..., 0] = x[..., 0] - x[..., 2] / 2  # top left x
    y[..., 1] = x[..., 1] - x[..., 3] / 2  # top left y
    y[..., 2] = x[..., 0] + x[..., 2] / 2  # bottom right x
    y[..., 3] = x[..., 1] + x[..., 3] / 2  # bottom right y
    return y


def _yolo_nms(boxes, scores, iou_threshold: float = 0.7) -> List[int]:
    """
    Perform Non-Maximum Suppression (NMS) on bounding boxes.

    This function applies NMS to remove overlapping bounding boxes, keeping only the most confident detections.

    :param boxes: Array of bounding boxes, each in the format [xmin, ymin, xmax, ymax].
    :type boxes: np.ndarray
    :param scores: Array of confidence scores for each bounding box.
    :type scores: np.ndarray
    :param iou_threshold: IoU threshold for considering boxes as overlapping. Default is 0.7.
    :type iou_threshold: float

    :return: List of indices of the boxes to keep after NMS.
    :rtype: List[int]

    :Example:

    >>> boxes = np.array([[0, 0, 10, 10], [1, 1, 11, 11], [20, 20, 30, 30]])
    >>> scores = np.array([0.9, 0.8, 0.7])
    >>> _yolo_nms(boxes, scores, 0.5)
    [0, 2]
    """
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    areas = (x2 - x1 + 1) * (y2 - y1 + 1)

    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)

        inter = w * h
        iou = inter / (areas[i] + areas[order[1:]] - inter)

        inds = np.where(iou <= iou_threshold)[0]
        order = order[inds + 1]

    return keep


def _image_preprocess(image: Image.Image, max_infer_size: Union[int, Tuple[int, int]] = 1216,
                      allow_dynamic: bool = False, align: int = 32):
    """
    Preprocess an input image for inference.

    This function resizes the image while maintaining its aspect ratio, and ensures
    the dimensions are multiples of 'align'.

    :param image: Input image to be preprocessed.
    :type image: Image.Image
    :param max_infer_size: Maximum size (width or height) of the processed image. Default is 1216.
    :type max_infer_size: int
    :param allow_dynamic: If True, allows dynamic resizing of the image while maintaining aspect ratio. Default is False.
    :type allow_dynamic: bool
    :param align: Value to align the image dimensions to. Default is 32.
    :type align: int

    :return: A tuple containing:
        - The preprocessed image
        - Original image dimensions (width, height)
        - New image dimensions (width, height)
    :rtype: tuple(Image.Image, Tuple[int, int], Tuple[int, int])

    :Example:

    >>> from PIL import Image
    >>> img = Image.new('RGB', (1000, 800))
    >>> processed_img, old_size, new_size = _image_preprocess(img)
    >>> print(old_size, new_size)
    (1000, 800) (1216, 992)
    """
    if isinstance(max_infer_size, int):
        max_infer_width, max_infer_height = max_infer_size, max_infer_size
    else:
        max_infer_width, max_infer_height = max_infer_size

    old_width, old_height = image.width, image.height
    new_width, new_height = old_width, old_height
    if allow_dynamic:
        r = min(max_infer_width / new_width, max_infer_height / new_height)
        if r < 1:
            new_width, new_height = new_width * r, new_height * r
        new_width = int(math.ceil(new_width / align) * align)
        new_height = int(math.ceil(new_height / align) * align)
    else:
        new_width, new_height = max_infer_width, max_infer_height

    image = image.resize((new_width, new_height))
    return image, (old_width, old_height), (new_width, new_height)


def _xy_postprocess(x, y, old_size: Tuple[float, float], new_size: Tuple[float, float]):
    """
    Convert coordinates from the preprocessed image size back to the original image size.

    :param x: X-coordinate in the preprocessed image.
    :type x: float
    :param y: Y-coordinate in the preprocessed image.
    :type y: float
    :param old_size: Original image dimensions (width, height).
    :type old_size: Tuple[float, float]
    :param new_size: Preprocessed image dimensions (width, height).
    :type new_size: Tuple[float, float]

    :return: Adjusted (x, y) coordinates for the original image size.
    :rtype: Tuple[int, int]

    :Example:

    >>> _xy_postprocess(100, 100, (1000, 800), (1216, 992))
    (82, 80)
    """
    old_width, old_height = old_size
    new_width, new_height = new_size
    x, y = x / new_width * old_width, y / new_height * old_height
    x = int(np.clip(x, a_min=0, a_max=old_width).round())
    y = int(np.clip(y, a_min=0, a_max=old_height).round())
    return x, y


def _end2end_postprocess(output, conf_threshold: float, iou_threshold: float,
                         old_size: Tuple[float, float], new_size: Tuple[float, float], labels: List[str]) \
        -> List[Tuple[Tuple[int, int, int, int], str, float]]:
    """
    Post-process the output of an end-to-end object detection model.

    This function filters detections based on confidence, applies non-maximum suppression,
    and transforms coordinates back to the original image size.

    :param output: Raw output from the end-to-end object detection model.
    :type output: np.ndarray
    :param conf_threshold: Confidence threshold for filtering detections.
    :type conf_threshold: float
    :param iou_threshold: IoU threshold for non-maximum suppression (not used in this function).
    :type iou_threshold: float
    :param old_size: Original image dimensions (width, height).
    :type old_size: Tuple[float, float]
    :param new_size: Preprocessed image dimensions (width, height).
    :type new_size: Tuple[float, float]
    :param labels: List of class labels.
    :type labels: List[str]

    :return: List of detections, each in the format ((x0, y0, x1, y1), label, confidence).
    :rtype: List[Tuple[Tuple[int, int, int, int], str, float]]

    :raises AssertionError: If the output shape is not as expected.
    """
    assert output.shape[-1] == 6
    _ = iou_threshold  # actually the iou_threshold has not been supplied to end2end post-processing
    detections = []
    output = output[output[:, 4] > conf_threshold]
    selected_idx = _yolo_nms(output[:, :4], output[:, 4])
    for x0, y0, x1, y1, score, cls in output[selected_idx]:
        x0, y0 = _xy_postprocess(x0, y0, old_size, new_size)
        x1, y1 = _xy_postprocess(x1, y1, old_size, new_size)
        detections.append(((x0, y0, x1, y1), labels[int(cls.item())], float(score)))

    return detections


def _nms_postprocess(output, conf_threshold: float, iou_threshold: float,
                     old_size: Tuple[float, float], new_size: Tuple[float, float], labels: List[str]) \
        -> List[Tuple[Tuple[int, int, int, int], str, float]]:
    """
    Post-process the output of an NMS-based object detection model.

    This function applies confidence thresholding, non-maximum suppression,
    and transforms coordinates back to the original image size.

    :param output: Raw output from the NMS-based object detection model.
    :type output: np.ndarray
    :param conf_threshold: Confidence threshold for filtering detections.
    :type conf_threshold: float
    :param iou_threshold: IoU threshold for non-maximum suppression.
    :type iou_threshold: float
    :param old_size: Original image dimensions (width, height).
    :type old_size: Tuple[float, float]
    :param new_size: Preprocessed image dimensions (width, height).
    :type new_size: Tuple[float, float]
    :param labels: List of class labels.
    :type labels: List[str]

    :return: List of detections, each in the format ((x0, y0, x1, y1), label, confidence).
    :rtype: List[Tuple[Tuple[int, int, int, int], str, float]]

    :raises AssertionError: If the output shape is not as expected.
    """
    assert output.shape[0] == 4 + len(labels)
    # the output should be like [4+cls, box_cnt]
    # cls means count of classes
    # box_cnt means count of bboxes
    max_scores = output[4:, :].max(axis=0)
    output = output[:, max_scores > conf_threshold].transpose(1, 0)
    boxes = output[:, :4]
    scores = output[:, 4:]
    filtered_max_scores = scores.max(axis=1)

    if not boxes.size:
        return []

    boxes = _yolo_xywh2xyxy(boxes)
    idx = _yolo_nms(boxes, filtered_max_scores, iou_threshold=iou_threshold)
    boxes, scores = boxes[idx], scores[idx]

    detections = []
    for box, score in zip(boxes, scores):
        x0, y0 = _xy_postprocess(box[0], box[1], old_size, new_size)
        x1, y1 = _xy_postprocess(box[2], box[3], old_size, new_size)
        max_score_id = score.argmax()
        detections.append(((x0, y0, x1, y1), labels[max_score_id], float(score[max_score_id])))

    return detections


def _yolo_postprocess(output, conf_threshold: float, iou_threshold: float,
                      old_size: Tuple[float, float], new_size: Tuple[float, float], labels: List[str]) \
        -> List[Tuple[Tuple[int, int, int, int], str, float]]:
    """
    Post-process the raw output from the object detection model.

    This function applies confidence thresholding, non-maximum suppression, and
    converts the coordinates back to the original image size.

    :param output: Raw output from the object detection model.
    :type output: np.ndarray
    :param conf_threshold: Confidence threshold for filtering detections.
    :type conf_threshold: float
    :param iou_threshold: IoU threshold for non-maximum suppression.
    :type iou_threshold: float
    :param old_size: Original image dimensions (width, height).
    :type old_size: Tuple[float, float]
    :param new_size: Preprocessed image dimensions (width, height).
    :type new_size: Tuple[float, float]
    :param labels: List of class labels.
    :type labels: List[str]

    :return: List of detections, each in the format ((x0, y0, x1, y1), label, confidence).
    :rtype: List[tuple(tuple(int, int, int, int), str, float)]

    :Example:

    >>> output = np.array([[10, 10, 20, 20, 0.9, 0.1]])
    >>> _yolo_postprocess(output, 0.5, 0.5, (100, 100), (128, 128), ['cat', 'dog'])
    [((7, 7, 15, 15), 'cat', 0.9)]
    """
    if output.shape[-1] == 6:  # for end-to-end models like yolov10
        return _end2end_postprocess(
            output=output,
            conf_threshold=conf_threshold,
            iou_threshold=iou_threshold,
            old_size=old_size,
            new_size=new_size,
            labels=labels,
        )
    else:  # for nms-based models like yolov8
        return _nms_postprocess(
            output=output,
            conf_threshold=conf_threshold,
            iou_threshold=iou_threshold,
            old_size=old_size,
            new_size=new_size,
            labels=labels,
        )


def _rtdetr_postprocess(output, conf_threshold: float, iou_threshold: float,
                        old_size: Tuple[int, int], new_size: Tuple[int, int], labels: List[str]) \
        -> List[Tuple[Tuple[int, int, int, int], str, float]]:
    """
    Post-process the output from an RT-DETR (Real-Time DEtection TRansformer) model.

    This function handles the specific output format of RT-DETR models and applies
    the necessary post-processing steps.

    :param output: Raw output from the RT-DETR model.
    :type output: np.ndarray
    :param conf_threshold: Confidence threshold for filtering detections.
    :type conf_threshold: float
    :param iou_threshold: IoU threshold for non-maximum suppression.
    :type iou_threshold: float
    :param old_size: Original image dimensions (width, height).
    :type old_size: Tuple[int, int]
    :param new_size: Preprocessed image dimensions (width, height) (not used in this function).
    :type new_size: Tuple[int, int]
    :param labels: List of class labels.
    :type labels: List[str]

    :return: List of detections, each in the format ((x0, y0, x1, y1), label, confidence).
    :rtype: List[Tuple[Tuple[int, int, int, int], str, float]]

    :raises AssertionError: If the output shape is not as expected.
    """
    assert output.shape[-1] == 4 + len(labels)
    # the size rtdetr using is [0.0, 1.0]
    _ = new_size
    return _nms_postprocess(
        output=output.transpose(1, 0),
        conf_threshold=conf_threshold,
        iou_threshold=iou_threshold,
        old_size=old_size,
        new_size=(1.0, 1.0),
        labels=labels,
    )


def _safe_eval_names_str(names_str):
    """
    Safely evaluate the names string from model metadata.

    :param names_str: String representation of names dictionary.
    :type names_str: str
    :return: Dictionary of name mappings.
    :rtype: dict
    :raises RuntimeError: If an invalid key or value type is encountered.

    This function parses the names string from the model metadata, ensuring that
    only string and number literals are evaluated for safety.
    """
    node = ast.parse(names_str, mode='eval')
    result = {}
    # noinspection PyUnresolvedReferences
    for key, value in zip(node.body.keys, node.body.values):
        if isinstance(key, (ast.Str, ast.Num)):
            key = ast.literal_eval(key)
        else:
            raise RuntimeError(f"Invalid key type: {key!r}, this should be a bug, "
                               f"please open an issue to dghs-imgutils.")  # pragma: no cover

        if isinstance(value, (ast.Str, ast.Num)):
            value = ast.literal_eval(value)
        else:
            raise RuntimeError(f"Invalid value type: {value!r}, this should be a bug, "
                               f"please open an issue to dghs-imgutils.")  # pragma: no cover

        result[key] = value

    return result


class YOLOModel:
    """
    A class to manage YOLO models from a Hugging Face repository.

    This class handles the loading, caching, and inference of YOLO models.

    :param repo_id: The Hugging Face repository ID containing the YOLO models.
    :type repo_id: str
    :param hf_token: Optional Hugging Face authentication token.
    :type hf_token: Optional[str]

    :Example:

    >>> model = YOLOModel("username/repo_name")
    >>> image = Image.open("path/to/image.jpg")
    >>> detections = model.predict(image, "model_name")
    """

    def __init__(self, repo_id: str, hf_token: Optional[str] = None):
        """
        Initialize the YOLOModel.

        :param repo_id: The Hugging Face repository ID containing the YOLO models.
        :type repo_id: str
        :param hf_token: Optional Hugging Face authentication token.
        :type hf_token: Optional[str]
        """
        self.repo_id = repo_id
        self._model_names = None
        self._models = {}
        self._model_types = {}
        self._hf_token = hf_token
        self._global_lock = Lock()
        self._model_lock = Lock()

    def _get_hf_token(self) -> Optional[str]:
        """
        Get the Hugging Face token, either from the instance or environment variable.

        :return: Hugging Face token.
        :rtype: Optional[str]
        """
        return self._hf_token or os.environ.get('HF_TOKEN')

    @property
    def model_names(self) -> List[str]:
        """
        Get the list of available model names in the repository.

        :return: List of model names.
        :rtype: List[str]
        """
        with self._global_lock:
            if self._model_names is None:
                hf_fs = HfFileSystem(token=self._get_hf_token())
                self._model_names = [
                    hf_normpath(os.path.dirname(os.path.relpath(item, self.repo_id)))
                    for item in hf_fs.glob(hf_fs_path(
                        repo_id=self.repo_id,
                        repo_type='model',
                        filename='*/model.onnx',
                    ))
                ]

        return self._model_names

    def _check_model_name(self, model_name: str):
        """
        Check if the given model name is valid for this repository.

        :param model_name: Name of the model to check.
        :type model_name: str
        :raises ValueError: If the model name is not found in the repository.
        """
        if model_name not in self.model_names:
            raise ValueError(f'Unknown model {model_name!r} in model repository {self.repo_id!r}, '
                             f'models {self.model_names!r} are available.')

    def _open_model(self, model_name: str):
        """
        Open and cache a YOLO model.

        :param model_name: Name of the model to open.
        :type model_name: str
        :return: Tuple containing the ONNX model, maximum inference size, and labels.
        :rtype: tuple
        """
        with self._model_lock:
            if model_name not in self._models:
                self._check_model_name(model_name)
                model = open_onnx_model(hf_hub_download(
                    self.repo_id,
                    f'{model_name}/model.onnx',
                    token=self._get_hf_token(),
                ))
                model_metadata = model.get_modelmeta()
                if 'imgsz' in model_metadata.custom_metadata_map:
                    max_infer_size = tuple(json.loads(model_metadata.custom_metadata_map['imgsz']))
                    assert len(max_infer_size) == 2, f'imgsz should have 2 dims, but {max_infer_size!r} found.'
                else:
                    max_infer_size = 640
                names_map = _safe_eval_names_str(model_metadata.custom_metadata_map['names'])
                labels = [names_map[i] for i in range(len(names_map))]
                self._models[model_name] = (model, max_infer_size, labels)

        return self._models[model_name]

    def _get_model_type(self, model_name: str):
        with self._model_lock:
            if model_name not in self._model_types:
                hf_fs = get_hf_fs(hf_token=self._get_hf_token())
                fs_path = hf_fs_path(
                    repo_id=self.repo_id,
                    repo_type='model',
                    filename=f'{model_name}/model_type.json',
                    revision='main',
                )
                if hf_fs.exists(fs_path):
                    model_type = json.loads(hf_fs.read_text(fs_path))['model_type']
                else:
                    model_type = 'yolo'
                self._model_types[model_name] = model_type

        return self._model_types[model_name]

    def predict(self, image: ImageTyping, model_name: str,
                conf_threshold: float = 0.25, iou_threshold: float = 0.7,
                allow_dynamic: bool = False) \
            -> List[Tuple[Tuple[int, int, int, int], str, float]]:
        """
        Perform object detection on an image using the specified YOLO model.

        :param image: Input image for object detection.
        :type image: ImageTyping
        :param model_name: Name of the YOLO model to use.
        :type model_name: str
        :param conf_threshold: Confidence threshold for filtering detections. Default is 0.25.
        :type conf_threshold: float
        :param iou_threshold: IoU threshold for non-maximum suppression. Default is 0.7.
        :type iou_threshold: float
        :param allow_dynamic: If True, allows dynamic resizing of the image while maintaining aspect ratio.
                              Default is False.
        :type allow_dynamic: bool

        :return: List of detections, each in the format ((x0, y0, x1, y1), label, confidence).
        :rtype: List[Tuple[Tuple[int, int, int, int], str, float]]

        :Example:

        >>> model = YOLOModel("username/repo_name")
        >>> image = Image.open("path/to/image.jpg")
        >>> detections = model.predict(image, "model_name")
        >>> print(detections[0])  # First detection
        ((100, 200, 300, 400), 'person', 0.95)
        """
        model, max_infer_size, labels = self._open_model(model_name)
        image = load_image(image, mode='RGB')
        new_image, old_size, new_size = _image_preprocess(image, max_infer_size, allow_dynamic=allow_dynamic)
        data = rgb_encode(new_image)[None, ...]
        output, = model.run(['output0'], {'images': data})
        model_type = self._get_model_type(model_name=model_name)
        if model_type == 'yolo':
            return _yolo_postprocess(
                output=output[0],
                conf_threshold=conf_threshold,
                iou_threshold=iou_threshold,
                old_size=old_size,
                new_size=new_size,
                labels=labels
            )
        elif model_type == 'rtdetr':
            return _rtdetr_postprocess(
                output=output[0],
                conf_threshold=conf_threshold,
                iou_threshold=iou_threshold,
                old_size=old_size,
                new_size=new_size,
                labels=labels
            )
        else:
            raise ValueError(f'Unknown object detection model type - {model_type!r}.')  # pragma: no cover

    def clear(self):
        """
        Clear cached model and metadata.

        This method removes all cached models and their associated metadata from memory.
        It's useful for freeing up memory or ensuring that the latest versions of models are loaded.
        """
        self._models.clear()

    def make_ui(self, default_model_name: Optional[str] = None,
                default_conf_threshold: float = 0.25, default_iou_threshold: float = 0.7):
        """
        Create a Gradio-based user interface for object detection.

        This method sets up an interactive UI that allows users to upload images,
        select models, and adjust detection parameters. It uses the Gradio library
        to create the interface.

        :param default_model_name: The name of the default model to use.
                                   If None, the most recently updated model is selected.
        :type default_model_name: Optional[str]
        :param default_conf_threshold: Default confidence threshold for the UI. Default is 0.25.
        :type default_conf_threshold: float
        :param default_iou_threshold: Default IoU threshold for the UI. Default is 0.7.
        :type default_iou_threshold: float

        :raises ImportError: If Gradio is not installed in the environment.

        :Example:

        >>> model = YOLOModel("username/repo_name")
        >>> model.make_ui(default_model_name="yolov5s")
        """
        _check_gradio_env()
        model_list = self.model_names
        if not default_model_name:
            hf_client = get_hf_client(hf_token=self._get_hf_token())
            selected_model_name, selected_time = None, None
            for fileitem in hf_client.get_paths_info(
                    repo_id=self.repo_id,
                    repo_type='model',
                    paths=[f'{model_name}/model.onnx' for model_name in model_list],
                    expand=True,
            ):
                if not selected_time or fileitem.last_commit.date > selected_time:
                    selected_model_name = os.path.dirname(fileitem.path)
                    selected_time = fileitem.last_commit.date
            default_model_name = selected_model_name

        def _gr_detect(image: ImageTyping, model_name: str,
                       iou_threshold: float = 0.7, score_threshold: float = 0.25,
                       allow_dynamic: bool = False) \
                -> gr.AnnotatedImage:
            _, _, labels = self._open_model(model_name=model_name)
            _colors = list(map(str, rnd_colors(len(labels))))
            _color_map = dict(zip(labels, _colors))
            return gr.AnnotatedImage(
                value=(image, [
                    (_bbox_fix(bbox), label)
                    for bbox, label, _ in self.predict(
                        image=image,
                        model_name=model_name,
                        iou_threshold=iou_threshold,
                        conf_threshold=score_threshold,
                        allow_dynamic=allow_dynamic,
                    )
                ]),
                color_map=_color_map,
                label='Labeled',
            )

        with gr.Row():
            with gr.Column():
                gr_input_image = gr.Image(type='pil', label='Original Image')
                with gr.Row():
                    gr_model = gr.Dropdown(model_list, value=default_model_name, label='Model')
                    gr_allow_dynamic = gr.Checkbox(value=False, label='Allow Dynamic Size')
                with gr.Row():
                    gr_iou_threshold = gr.Slider(0.0, 1.0, default_iou_threshold, label='IOU Threshold')
                    gr_score_threshold = gr.Slider(0.0, 1.0, default_conf_threshold, label='Score Threshold')

                gr_submit = gr.Button(value='Submit', variant='primary')

            with gr.Column():
                gr_output_image = gr.AnnotatedImage(label="Labeled")

            gr_submit.click(
                _gr_detect,
                inputs=[
                    gr_input_image,
                    gr_model,
                    gr_iou_threshold,
                    gr_score_threshold,
                    gr_allow_dynamic,
                ],
                outputs=[gr_output_image],
            )

    def launch_demo(self, default_model_name: Optional[str] = None,
                    default_conf_threshold: float = 0.25, default_iou_threshold: float = 0.7,
                    server_name: Optional[str] = None, server_port: Optional[int] = None, **kwargs):
        """
        Launch a Gradio demo for object detection.

        This method creates and launches a Gradio demo that allows users to interactively
        perform object detection on uploaded images using the YOLO model.

        :param default_model_name: The name of the default model to use.
                                   If None, the most recently updated model is selected.
        :type default_model_name: Optional[str]
        :param default_conf_threshold: Default confidence threshold for the demo. Default is 0.25.
        :type default_conf_threshold: float
        :param default_iou_threshold: Default IoU threshold for the demo. Default is 0.7.
        :type default_iou_threshold: float
        :param server_name: The name of the server to run the demo on. Default is None.
        :type server_name: Optional[str]
        :param server_port: The port to run the demo on. Default is None.
        :type server_port: Optional[int]
        :param kwargs: Additional keyword arguments to pass to gr.Blocks.launch().

        :raises EnvironmentError: If Gradio is not installed in the environment.

        Example:
            >>> model = YOLOModel("username/repo_name")
            >>> model.launch_demo(default_model_name="yolov5s", server_name="0.0.0.0", server_port=7860)
        """
        _check_gradio_env()
        with gr.Blocks() as demo:
            with gr.Row():
                with gr.Column():
                    repo_url = hf_hub_repo_url(repo_id=self.repo_id, repo_type='model')
                    gr.HTML(f'<h2 style="text-align: center;">YOLO Demo For {self.repo_id}</h2>')
                    gr.Markdown(f'This is the quick demo for YOLO model [{self.repo_id}]({repo_url}). '
                                f'Powered by `dghs-imgutils`\'s quick demo module.')

            with gr.Row():
                self.make_ui(
                    default_model_name=default_model_name,
                    default_conf_threshold=default_conf_threshold,
                    default_iou_threshold=default_iou_threshold,
                )

        demo.launch(
            server_name=server_name,
            server_port=server_port,
            **kwargs,
        )


@ts_lru_cache()
def _open_models_for_repo_id(repo_id: str, hf_token: Optional[str] = None) -> YOLOModel:
    """
    Load and cache a YOLO model from a Hugging Face repository.

    This function uses the `lru_cache` decorator to cache the loaded models,
    improving performance for repeated calls with the same repository ID.

    :param repo_id: The Hugging Face repository ID for the YOLO model.
    :type repo_id: str
    :param hf_token: Optional Hugging Face authentication token.
    :type hf_token: Optional[str]

    :return: The loaded YOLO model.
    :rtype: YOLOModel

    :raises Exception: If there's an error loading the model from the repository.

    Usage:
        >>> model = _open_models_for_repo_id("yolov5/yolov5s")
        >>> # Subsequent calls with the same repo_id will return the cached model
        >>> same_model = _open_models_for_repo_id("yolov5/yolov5s")
    """
    return YOLOModel(repo_id, hf_token=hf_token)


def yolo_predict(image: ImageTyping, repo_id: str, model_name: str,
                 conf_threshold: float = 0.25, iou_threshold: float = 0.7,
                 hf_token: Optional[str] = None, **kwargs) \
        -> List[Tuple[Tuple[int, int, int, int], str, float]]:
    """
    Perform object detection on an image using a YOLO model from a Hugging Face repository.

    This function is a high-level wrapper around the YOLOModel class, providing a simple
    interface for object detection without needing to explicitly manage model instances.

    :param image: Input image for object detection.
    :type image: ImageTyping
    :param repo_id: The Hugging Face repository ID containing the YOLO models.
    :type repo_id: str
    :param model_name: Name of the YOLO model to use.
    :type model_name: str
    :param conf_threshold: Confidence threshold for filtering detections. Default is 0.25.
    :type conf_threshold: float
    :param iou_threshold: IoU threshold for non-maximum suppression. Default is 0.7.
    :type iou_threshold: float
    :param hf_token: Optional Hugging Face authentication token.
    :type hf_token: Optional[str]

    :return: List of detections, each in the format ((x0, y0, x1, y1), label, confidence).
    :rtype: List[Tuple[Tuple[int, int, int, int], str, float]]

    :Example:

    >>> from PIL import Image
    >>> image = Image.open("path/to/image.jpg")
    >>> detections = yolo_predict(image, "username/repo_name", "model_name")
    >>> print(detections[0])  # First detection
    ((100, 200, 300, 400), 'person', 0.95)
    """
    return _open_models_for_repo_id(repo_id, hf_token=hf_token).predict(
        image=image,
        model_name=model_name,
        conf_threshold=conf_threshold,
        iou_threshold=iou_threshold,
        **kwargs,
    )



"""
Overview:
    A model for classifying anime real images into 2 classes (``anime``, ``real``).

    The following are sample images for testing.

    .. image:: real.plot.py.svg
        :align: center

    This is an overall benchmark of all the real classification models:

    .. image:: real_benchmark.plot.py.svg
        :align: center

    The models are hosted on
    `huggingface - deepghs/anime_real_cls <https://huggingface.co/deepghs/anime_real_cls>`_.
"""
from typing import Tuple, Dict

from ..data import ImageTyping
from ..generic import classify_predict, classify_predict_score

__all__ = [
    'anime_real_score',
    'anime_real',
]

_DEFAULT_MODEL_NAME = 'mobilenetv3_v1.4_dist'
_REPO_ID = 'deepghs/anime_real_cls'


def anime_real_score(image: ImageTyping, model_name: str = _DEFAULT_MODEL_NAME) -> Dict[str, float]:
    """
    Get the scores for different types in an anime real.

    :param image: The input image.
    :type image: ImageTyping
    :param model_name: The model name. Default is 'mobilenetv3_v1.2_dist'.
    :type model_name: str
    :return: A dictionary with type scores.
    :rtype: Dict[str, float]

    Examples::
        >>> from imgutils.validate import anime_real_score
        >>>
        >>> anime_real_score('real/anime/1.jpg')
        {'anime': 0.9999716281890869, 'real': 2.8398366339388303e-05}
        >>> anime_real_score('real/anime/2.jpg')
        {'anime': 0.9992202520370483, 'real': 0.0007797438884153962}
        >>> anime_real_score('real/anime/3.jpg')
        {'anime': 0.9999709129333496, 'real': 2.905452492996119e-05}
        >>> anime_real_score('real/anime/4.jpg')
        {'anime': 0.9999765157699585, 'real': 2.3499671442550607e-05}
        >>> anime_real_score('real/anime/5.jpg')
        {'anime': 0.9994087219238281, 'real': 0.0005913018831051886}
        >>> anime_real_score('real/anime/6.jpg')
        {'anime': 0.9999759197235107, 'real': 2.4061362637439743e-05}
        >>> anime_real_score('real/anime/7.jpg')
        {'anime': 0.9999052286148071, 'real': 9.475799015490338e-05}
        >>> anime_real_score('real/anime/8.jpg')
        {'anime': 0.9999759197235107, 'real': 2.403173675702419e-05}
        >>> anime_real_score('real/real/9.jpg')
        {'anime': 1.5848207794988411e-06, 'real': 0.9999984502792358}
        >>> anime_real_score('real/real/10.jpg')
        {'anime': 0.0010207017185166478, 'real': 0.9989792704582214}
        >>> anime_real_score('real/real/11.jpg')
        {'anime': 2.2124368115328252e-06, 'real': 0.9999977350234985}
        >>> anime_real_score('real/real/12.jpg')
        {'anime': 1.6512358342879452e-05, 'real': 0.9999834299087524}
        >>> anime_real_score('real/real/13.jpg')
        {'anime': 6.359853614412714e-06, 'real': 0.9999936819076538}
        >>> anime_real_score('real/real/14.jpg')
        {'anime': 1.600314317329321e-05, 'real': 0.9999840259552002}
        >>> anime_real_score('real/real/15.jpg')
        {'anime': 1.5589323083986528e-05, 'real': 0.9999843835830688}
        >>> anime_real_score('real/real/16.jpg')
        {'anime': 1.5513256585109048e-05, 'real': 0.9999845027923584}
    """
    return classify_predict_score(image, _REPO_ID, model_name)


def anime_real(image: ImageTyping, model_name: str = _DEFAULT_MODEL_NAME) -> Tuple[str, float]:
    """
    Get the primary anime real type and its score.

    :param image: The input image.
    :type image: ImageTyping
    :param model_name: The model name. Default is 'mobilenetv3_v1.2_dist'.
    :type model_name: str
    :return: A tuple with the primary type and its score.
    :rtype: Tuple[str, float]

    Examples::
        >>> from imgutils.validate import anime_real
        >>>
        >>> anime_real('real/anime/1.jpg')
        ('anime', 0.9999716281890869)
        >>> anime_real('real/anime/2.jpg')
        ('anime', 0.9992202520370483)
        >>> anime_real('real/anime/3.jpg')
        ('anime', 0.9999709129333496)
        >>> anime_real('real/anime/4.jpg')
        ('anime', 0.9999765157699585)
        >>> anime_real('real/anime/5.jpg')
        ('anime', 0.9994087219238281)
        >>> anime_real('real/anime/6.jpg')
        ('anime', 0.9999759197235107)
        >>> anime_real('real/anime/7.jpg')
        ('anime', 0.9999052286148071)
        >>> anime_real('real/anime/8.jpg')
        ('anime', 0.9999759197235107)
        >>> anime_real('real/real/9.jpg')
        ('real', 0.9999984502792358)
        >>> anime_real('real/real/10.jpg')
        ('real', 0.9989792704582214)
        >>> anime_real('real/real/11.jpg')
        ('real', 0.9999977350234985)
        >>> anime_real('real/real/12.jpg')
        ('real', 0.9999834299087524)
        >>> anime_real('real/real/13.jpg')
        ('real', 0.9999936819076538)
        >>> anime_real('real/real/14.jpg')
        ('real', 0.9999840259552002)
        >>> anime_real('real/real/15.jpg')
        ('real', 0.9999843835830688)
        >>> anime_real('real/real/16.jpg')
        ('real', 0.9999845027923584)
    """
    return classify_predict(image, _REPO_ID, model_name)



"""
Overview:
    Tagging utils based on deepdanbooru.

    .. warning::
        Due to the usage of an outdated model and training data in deepdanbooru,
        its performance is limited, and it is **not suitable for use as the main tagging model anymore**.
        The integration of this model within the present project serves only as a baseline for comparison,
        and it is advisable to avoid using this model extensively in practical applications.
"""
from typing import Tuple, List

import numpy as np
import pandas as pd
from PIL import Image
from huggingface_hub import hf_hub_download

from .overlap import drop_overlap_tags
from ..data import ImageTyping, load_image
from ..utils import open_onnx_model, ts_lru_cache


@ts_lru_cache()
def _get_deepdanbooru_labels():
    csv_file = hf_hub_download('deepghs/imgutils-models', 'deepdanbooru/deepdanbooru_tags.csv')
    df = pd.read_csv(csv_file)

    tag_names = df["name"].tolist()
    tag_real_names = df['real_name'].tolist()
    rating_indexes = list(np.where(df["category"] == 9)[0])
    general_indexes = list(np.where(df["category"] == 0)[0])
    character_indexes = list(np.where(df["category"] == 4)[0])
    return tag_names, tag_real_names, \
        rating_indexes, general_indexes, character_indexes


@ts_lru_cache()
def _get_deepdanbooru_model():
    return open_onnx_model(hf_hub_download(
        'deepghs/imgutils-models',
        'deepdanbooru/deepdanbooru.onnx',
    ))


def _image_preprocess(image: Image.Image) -> np.ndarray:
    o_width, o_height = image.size
    scale = 512.0 / max(o_width, o_height)
    f_width, f_height = map(lambda x: int(x * scale), (o_width, o_height))
    image = image.resize((f_width, f_height))

    data = np.asarray(image).astype(np.float32) / 255  # H x W x C
    height_pad_left = (512 - f_height) // 2
    height_pad_right = 512 - f_height - height_pad_left
    width_pad_left = (512 - f_width) // 2
    width_pad_right = 512 - f_width - width_pad_left
    data = np.pad(data, ((height_pad_left, height_pad_right), (width_pad_left, width_pad_right), (0, 0)),
                  mode='constant', constant_values=0.0)

    assert data.shape == (512, 512, 3), f'Shape (512, 512, 3) expected, but {data.shape!r} found.'
    return data.reshape((1, 512, 512, 3))  # B x H x W x C


def get_deepdanbooru_tags(image: ImageTyping, use_real_name: bool = False,
                          general_threshold: float = 0.5, character_threshold: float = 0.5,
                          drop_overlap: bool = False):
    """
    Overview:
        Get tags for anime image based on ``deepdanbooru`` model.

    :param image: Image to tagging.
    :param use_real_name: Use real name on danbooru. Due to the renaming and redirection of many tags
        on the Danbooru website after the training of ``deepdanbooru``,
        it may be necessary to use the latest tag names in some application scenarios.
        The default value of ``False`` indicates the use of the original tag names.
    :param general_threshold: Threshold for default tags, default is ``0.35``.
    :param character_threshold: Threshold for character tags, default is ``0.85``.
    :param drop_overlap: Drop overlap tags or not, default is ``False``.
    :return: Tagging results for levels, features and characters.

    Example:
        Here are some images for example

        .. image:: tagging_demo.plot.py.svg
           :align: center

        >>> from imgutils.tagging import get_deepdanbooru_tags
        >>>
        >>> rating, features, chars = get_deepdanbooru_tags('skadi.jpg')
        >>> rating
        {'rating:safe': 0.9897817373275757, 'rating:questionable': 0.010265946388244629, 'rating:explicit': 5.2809715270996094e-05}
        >>> features
        {'1girl': 0.9939777851104736, 'bangs': 0.5032387375831604, 'black_border': 0.9943548440933228, 'black_gloves': 0.5011609792709351, 'blue_sky': 0.6877802610397339, 'blush': 0.5543792843818665, 'breasts': 0.8268730640411377, 'cloud': 0.8504303693771362, 'cowboy_shot': 0.6008237600326538, 'crop_top': 0.6635787487030029, 'day': 0.8496965765953064, 'gloves': 0.6107005476951599, 'hair_between_eyes': 0.668294370174408, 'holding': 0.5619469285011292, 'holding_baseball_bat': 0.5141720771789551, 'letterboxed': 1.0, 'long_hair': 0.9884189963340759, 'looking_at_viewer': 0.5673105120658875, 'midriff': 0.6290556192398071, 'navel': 0.9631235003471375, 'no_hat': 0.7978747487068176, 'no_headwear': 0.7577926516532898, 'outdoors': 0.7118550539016724, 'parted_lips': 0.5452839136123657, 'pillarboxed': 0.9841411709785461, 'red_eyes': 0.958786129951477, 'shirt': 0.6720131039619446, 'short_sleeves': 0.7077711820602417, 'silver_hair': 0.6673924326896667, 'sky': 0.8709812760353088, 'solo': 0.9614333510398865, 'sportswear': 0.7786177396774292, 'standing': 0.6842771172523499, 'sweat': 0.9076308012008667, 'thighs': 0.580970823764801}
        >>> chars
        {'skadi_(arknights)': 0.9633345007896423}
        >>>
        >>> rating, features, chars = get_deepdanbooru_tags('hutao.jpg')
        >>> rating
        {'rating:safe': 0.9988503456115723, 'rating:questionable': 0.001651763916015625, 'rating:explicit': 0.00012505054473876953}
        >>> features
        {'1girl': 0.9829280972480774, ':p': 0.894218385219574, 'ahoge': 0.8733789920806885, 'backpack': 0.6322951316833496, 'bag': 0.9987058639526367, 'bag_charm': 0.9754379987716675, 'bangs': 0.6810564994812012, 'black_border': 0.9708781838417053, 'blush': 0.6356008052825928, 'bow': 0.5633733868598938, 'brick_wall': 0.5315935611724854, 'brown_hair': 0.9397273659706116, 'building': 0.9229896664619446, 'charm_(object)': 0.9006357789039612, 'city': 0.9020784497261047, 'cityscape': 0.9547432661056519, 'cowboy_shot': 0.5296419262886047, 'flower': 0.8253412246704102, 'hair_between_eyes': 0.5619839429855347, 'hair_flower': 0.8277763724327087, 'hair_ornament': 0.9356368780136108, 'hair_ribbon': 0.5288072824478149, 'jacket': 0.6336134076118469, 'letterboxed': 1.0, 'long_hair': 0.9703260064125061, 'looking_at_viewer': 0.8188960552215576, 'phone_screen': 0.9579574465751648, 'pillarboxed': 0.9954615235328674, 'plaid': 0.9725285172462463, 'plaid_skirt': 0.9638455510139465, 'pleated_skirt': 0.7226815819740295, 'red_eyes': 0.5321241021156311, 'red_nails': 0.5493080615997314, 'school_bag': 0.9863407611846924, 'school_uniform': 0.6794284582138062, 'shirt': 0.5062428116798401, 'shoulder_bag': 0.9325523972511292, 'skirt': 0.92237788438797, 'skyscraper': 0.7728171348571777, 'sleeves_past_wrists': 0.7257086038589478, 'smile': 0.5357837080955505, 'solo': 0.6939404010772705, 'thighhighs': 0.7054293155670166, 'tongue': 0.9990814924240112, 'tongue_out': 0.9992498755455017, 'twintails': 0.5012534260749817, 'very_long_hair': 0.7461410164833069}
        >>> chars
        {}
    """
    session = _get_deepdanbooru_model()
    _image_data = _image_preprocess(load_image(image, mode='RGB'))

    input_name = session.get_inputs()[0].name
    output_names = [output.name for output in session.get_outputs()]
    probs = session.run(output_names, {input_name: _image_data})[0]

    tag_names, tag_real_names, rating_indexes, general_indexes, character_indexes = _get_deepdanbooru_labels()
    labels: List[Tuple[str, float]] = list(zip(
        tag_real_names if use_real_name else tag_names,
        probs[0].astype(float).tolist(),
    ))

    # First 4 labels are actually ratings: pick one with argmax
    ratings_names = [labels[i] for i in rating_indexes]
    rating = dict(ratings_names)

    # Then we have general tags: pick anywhere prediction confidence > threshold
    general_names = [labels[i] for i in general_indexes]
    general_res = [x for x in general_names if x[1] > general_threshold]
    general_res = dict(general_res)
    if drop_overlap:
        general_res = drop_overlap_tags(general_res)

    # Everything else is characters: pick anywhere prediction confidence > threshold
    character_names = [labels[i] for i in character_indexes]
    character_res = [x for x in character_names if x[1] > character_threshold]
    character_res = dict(character_res)

    return rating, general_res, character_res



"""
This module provides functions for calculating similarities between bounding boxes and detections.

It includes functions to calculate Intersection over Union (IoU) for individual bounding boxes,
compute similarities between lists of bounding boxes, and compare detections with labels.
The module is designed to work with various types of bounding box representations and
offers different modes for aggregating similarity scores.

Key components:

- calculate_iou: Computes IoU between two bounding boxes
- bboxes_similarity: Calculates similarities between two lists of bounding boxes
- detection_similarity: Compares two lists of detections, considering both bounding boxes and labels

This module is particularly useful for tasks involving object detection, 
image segmentation, and evaluation of detection algorithms.
"""

from typing import List, Literal, Union

import numpy as np

from .base import BBoxTyping, BBoxWithScoreAndLabel


def calculate_iou(box1: BBoxTyping, box2: BBoxTyping) -> float:
    """
    Calculate the Intersection over Union (IoU) between two bounding boxes.

    :param box1: The first bounding box, represented as (x1, y1, x2, y2).
    :type box1: BBoxTyping
    :param box2: The second bounding box, represented as (x1, y1, x2, y2).
    :type box2: BBoxTyping
    :return: The IoU value between the two bounding boxes.
    :rtype: float

    This function computes the IoU, which is a measure of the overlap between two bounding boxes.
    The IoU is calculated as the area of intersection divided by the area of union of the two boxes.

    Example::
        >>> box1 = (0, 0, 2, 2)
        >>> box2 = (1, 1, 3, 3)
        >>> iou = calculate_iou(box1, box2)
        >>> print(f"IoU: {iou:.4f}")
        IoU: 0.1429
    """
    x1 = max(box1[0], box2[0])
    y1 = max(box1[1], box2[1])
    x2 = min(box1[2], box2[2])
    y2 = min(box1[3], box2[3])

    intersection = max(0.0, x2 - x1) * max(0.0, y2 - y1)
    area1 = (box1[2] - box1[0]) * (box1[3] - box1[1])
    area2 = (box2[2] - box2[0]) * (box2[3] - box2[1])

    iou = intersection / (area1 + area2 - intersection + 1e-6)
    return float(iou)


def bboxes_similarity(bboxes1: List[BBoxTyping], bboxes2: List[BBoxTyping],
                      mode: Literal['max', 'mean', 'raw'] = 'mean') -> Union[float, List[float]]:
    """
    Calculate the similarity between two lists of bounding boxes.

    :param bboxes1: First list of bounding boxes.
    :type bboxes1: List[BBoxTyping]
    :param bboxes2: Second list of bounding boxes.
    :type bboxes2: List[BBoxTyping]
    :param mode: The mode for calculating similarity. Options are 'max', 'mean', or 'raw'. Defaults to 'mean'.
    :type mode: Literal['max', 'mean', 'raw']
    :return: The similarity score or list of scores, depending on the mode.
    :rtype: Union[float, List[float]]
    :raises ValueError: If the lengths of bboxes1 and bboxes2 do not match, or if an unknown mode is specified.

    This function computes the similarity between two lists of bounding boxes using the Hungarian algorithm
    to find the optimal assignment. It then returns the similarity based on the specified mode:

    - ``max``: Returns the maximum IoU among all matched pairs.
    - ``mean``: Returns the average IoU of all matched pairs.
    - ``raw``: Returns a list of IoU values for all matched pairs.

    Example::
        >>> bboxes1 = [(0, 0, 2, 2), (3, 3, 5, 5)]
        >>> bboxes2 = [(1, 1, 3, 3), (4, 4, 6, 6)]
        >>> similarity = bboxes_similarity(bboxes1, bboxes2, mode='mean')
        >>> print(f"Mean similarity: {similarity:.4f}")
        Mean similarity: 0.1429
    """
    if len(bboxes1) != len(bboxes2):
        raise ValueError(f'Length of bboxes lists not match - {len(bboxes1)} vs {len(bboxes2)}.')

    n = len(bboxes1)
    iou_matrix = np.zeros((n, n))
    for i in range(n):
        for j in range(n):
            iou_matrix[i, j] = calculate_iou(bboxes1[i], bboxes2[j])

    # import here for faster launching speed
    from scipy.optimize import linear_sum_assignment
    row_ind, col_ind = linear_sum_assignment(-iou_matrix)
    similarities = iou_matrix[row_ind, col_ind]
    if mode == 'max':
        return float(similarities.max())
    elif mode == 'mean':
        return float(similarities.mean())
    elif mode == 'raw':
        return similarities.tolist()
    else:
        raise ValueError(f'Unknown similarity mode for bboxes - {mode!r}.')


def detection_similarity(detect1: List[BBoxWithScoreAndLabel], detect2: List[BBoxWithScoreAndLabel],
                         mode: Literal['max', 'mean', 'raw'] = 'mean') -> Union[float, List[float]]:
    """
    Calculate the similarity between two lists of detections, considering both bounding boxes and labels.

    :param detect1: First list of detections, each containing a bounding box, label, and score.
    :type detect1: List[BBoxWithScoreAndLabel]
    :param detect2: Second list of detections, each containing a bounding box, label, and score.
    :type detect2: List[BBoxWithScoreAndLabel]
    :param mode: The mode for calculating similarity. Options are 'max', 'mean', or 'raw'. Defaults to 'mean'.
    :type mode: Literal['max', 'mean', 'raw']
    :return: The similarity score or list of scores, depending on the mode.
    :rtype: Union[float, List[float]]
    :raises ValueError: If the number of bounding boxes for any label doesn't match between detect1 and detect2,
                        or if an unknown mode is specified.

    This function compares two lists of detections by:

    1. Grouping detections by their labels.
    2. For each label, calculating the similarity between the corresponding bounding boxes.
    3. Aggregating the similarities based on the specified mode.

    The function ensures that for each label, the number of bounding boxes matches between detect1 and detect2.

    Example::
        >>> detect1 = [((0, 0, 2, 2), 'car', 0.9), ((3, 3, 5, 5), 'person', 0.8)]
        >>> detect2 = [((1, 1, 3, 3), 'car', 0.85), ((4, 4, 6, 6), 'person', 0.75)]
        >>> similarity = detection_similarity(detect1, detect2, mode='mean')
        >>> print(f"Mean detection similarity: {similarity:.4f}")
        Mean detection similarity: 0.1429
    """
    labels = sorted({*(l for _, l, _ in detect1), *(l for _, l, _ in detect2)})
    sims = []
    for current_label in labels:
        bboxes1 = [bbox for bbox, label, _ in detect1 if label == current_label]
        bboxes2 = [bbox for bbox, label, _ in detect2 if label == current_label]

        if len(bboxes1) != len(bboxes2):
            raise ValueError(f'Length of bboxes not match on label {current_label!r}'
                             f' - {len(bboxes1)} vs {len(bboxes2)}.')

        sims.extend(bboxes_similarity(
            bboxes1=bboxes1,
            bboxes2=bboxes2,
            mode='raw',
        ))

    sims = np.array(sims)
    if mode == 'max':
        return float(sims.max())
    elif mode == 'mean':
        return float(sims.mean())
    elif mode == 'raw':
        return sims.tolist()
    else:
        raise ValueError(f'Unknown similarity mode for bboxes - {mode!r}.')



from typing import List

import cv2
import numpy as np
import pyclipper
from huggingface_hub import hf_hub_download, HfFileSystem
from shapely import Polygon

from ..data import ImageTyping, load_image
from ..utils import open_onnx_model, ts_lru_cache

_MIN_SIZE = 3
_HF_CLIENT = HfFileSystem()
_REPOSITORY = 'deepghs/paddleocr'


@ts_lru_cache()
def _open_ocr_detection_model(model):
    return open_onnx_model(hf_hub_download(
        _REPOSITORY,
        f'det/{model}/model.onnx',
    ))


def _box_score_fast(bitmap, _box):
    h, w = bitmap.shape[:2]
    box = _box.copy()
    xmin = np.clip(np.floor(box[:, 0].min()).astype("int32"), 0, w - 1)
    xmax = np.clip(np.ceil(box[:, 0].max()).astype("int32"), 0, w - 1)
    ymin = np.clip(np.floor(box[:, 1].min()).astype("int32"), 0, h - 1)
    ymax = np.clip(np.ceil(box[:, 1].max()).astype("int32"), 0, h - 1)

    mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)
    box[:, 0] = box[:, 0] - xmin
    box[:, 1] = box[:, 1] - ymin
    # noinspection PyTypeChecker
    cv2.fillPoly(mask, box.reshape(1, -1, 2).astype("int32"), 1)
    return cv2.mean(bitmap[ymin:ymax + 1, xmin:xmax + 1], mask)[0]


def _unclip(box, unclip_ratio):
    poly = Polygon(box)
    distance = poly.area * unclip_ratio / poly.length
    offset = pyclipper.PyclipperOffset()
    offset.AddPath(box, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
    expanded = np.array(offset.Execute(distance))
    return expanded


def _get_mini_boxes(contour):
    bounding_box = cv2.minAreaRect(contour)
    points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])

    if points[1][1] > points[0][1]:
        index_1 = 0
        index_4 = 1
    else:
        index_1 = 1
        index_4 = 0

    if points[3][1] > points[2][1]:
        index_2 = 2
        index_3 = 3
    else:
        index_2 = 3
        index_3 = 2

    box = [
        points[index_1], points[index_2], points[index_3], points[index_4]
    ]
    return box, min(bounding_box[1])


def _boxes_from_bitmap(pred, _bitmap, dest_width, dest_height,
                       box_threshold=0.7, max_candidates=1000, unclip_ratio=2.0):
    bitmap = _bitmap
    height, width = bitmap.shape

    outs = cv2.findContours((bitmap * 255).astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
    if len(outs) == 3:
        img, contours, _ = outs[0], outs[1], outs[2]
    elif len(outs) == 2:
        contours, _ = outs[0], outs[1]

    # noinspection PyUnboundLocalVariable
    num_contours = min(len(contours), max_candidates)

    boxes = []
    scores = []
    for index in range(num_contours):
        contour = contours[index]
        points, sside = _get_mini_boxes(contour)
        if sside < _MIN_SIZE:
            continue
        points = np.array(points)
        score = _box_score_fast(pred, points.reshape(-1, 2))
        if box_threshold > score:
            continue

        box = _unclip(points, unclip_ratio).reshape(-1, 1, 2)
        box, sside = _get_mini_boxes(box)
        if sside < _MIN_SIZE + 2:
            continue
        box = np.array(box)

        box[:, 0] = np.clip(np.round(box[:, 0] / width * dest_width), 0, dest_width)
        box[:, 1] = np.clip(np.round(box[:, 1] / height * dest_height), 0, dest_height)
        boxes.append(box.astype("int32"))
        scores.append(score)
    return np.array(boxes, dtype="int32"), scores


def _normalize(data, mean=(0.48145466, 0.4578275, 0.40821073), std=(0.26862954, 0.26130258, 0.27577711)):
    mean, std = np.asarray(mean), np.asarray(std)
    return (data - mean[None, :, None, None]) / std[None, :, None, None]


_ALIGN = 64


def _get_text_points(image: ImageTyping, model: str = 'ch_PP-OCRv4_det',
                     heat_threshold: float = 0.3, box_threshold: float = 0.7,
                     max_candidates: int = 1000, unclip_ratio: float = 2.0):
    origin_width, origin_height = width, height = image.size
    if width % _ALIGN != 0:
        width += (_ALIGN - width % _ALIGN)
    if height % _ALIGN != 0:
        height += (_ALIGN - height % _ALIGN)

    input_ = np.array(image).transpose((2, 0, 1)).astype(np.float32) / 255.0
    # noinspection PyTypeChecker
    input_ = np.pad(input_[None, ...], ((0, 0), (0, 0), (0, height - origin_height), (0, width - origin_width)))

    _ort_session = _open_ocr_detection_model(model)

    input_ = _normalize(input_).astype(np.float32)
    _input_name = _ort_session.get_inputs()[0].name
    _output_name = _ort_session.get_outputs()[0].name
    output_, = _ort_session.run([_output_name], {_input_name: input_})
    heatmap = output_[0][0]
    heatmap = heatmap[:origin_height, :origin_width]

    retval = []
    for points, score in zip(*_boxes_from_bitmap(
            heatmap, heatmap >= heat_threshold, origin_width, origin_height,
            box_threshold, max_candidates, unclip_ratio,
    )):
        retval.append((points, score))
    return retval


def _detect_text(image: ImageTyping, model: str = 'ch_PP-OCRv4_det',
                 heat_threshold: float = 0.3, box_threshold: float = 0.7,
                 max_candidates: int = 1000, unclip_ratio: float = 2.0):
    image = load_image(image, force_background='white', mode='RGB')
    retval = []
    for points, score in _get_text_points(image, model, heat_threshold, box_threshold, max_candidates, unclip_ratio):
        x0, y0 = points[:, 0].min(), points[:, 1].min()
        x1, y1 = points[:, 0].max(), points[:, 1].max()
        retval.append(((x0.item(), y0.item(), x1.item(), y1.item()), 'text', score))

    return retval


@ts_lru_cache()
def _list_det_models() -> List[str]:
    retval = []
    repo_segment_cnt = len(_REPOSITORY.split('/'))
    for item in _HF_CLIENT.glob(f'{_REPOSITORY}/det/*/model.onnx'):
        retval.append(item.split('/')[repo_segment_cnt:][1])
    return retval



"""
Overview:
    This module provides functionality for detecting nudity in images using the NudeNet model.
    
    The module includes functions for preprocessing images, running the NudeNet YOLO model,
    applying non-maximum suppression (NMS), and postprocessing the results. It utilizes
    ONNX models hosted on `deepghs/nudenet_onnx <https://huggingface.co/deepghs/nudenet_onnx>`_
    for efficient inference. The original project is
    `notAI-tech/NudeNet <https://github.com/notAI-tech/NudeNet>`_.
    
    .. collapse:: Overview of NudeNet Detect (NSFW Warning!!!)

        .. image:: nudenet_detect_demo.plot.py.svg
            :align: center
    
    The main function :func:`detect_with_nudenet` can be used to perform nudity detection on
    given images, returning a list of bounding boxes, labels, and confidence scores.
    
    This is an overall benchmark of all the nudenet models:

    .. image:: nudenet_detect_benchmark.plot.py.svg
        :align: center

    .. note::

        Here is a detailed list of labels from the NudeNet detection model and their respective meanings:

        .. list-table::
           :widths: 25 75
           :header-rows: 1

           * - Label
             - Description
           * - FEMALE_GENITALIA_COVERED
             - Detects covered female genitalia in the image.
           * - FACE_FEMALE
             - Detects the face of a female in the image.
           * - BUTTOCKS_EXPOSED
             - Detects exposed buttocks in the image.
           * - FEMALE_BREAST_EXPOSED
             - Detects exposed female breasts in the image.
           * - FEMALE_GENITALIA_EXPOSED
             - Detects exposed female genitalia in the image.
           * - MALE_BREAST_EXPOSED
             - Detects exposed male breasts in the image.
           * - ANUS_EXPOSED
             - Detects exposed anus in the image.
           * - FEET_EXPOSED
             - Detects exposed feet in the image.
           * - BELLY_COVERED
             - Detects a covered belly in the image.
           * - FEET_COVERED
             - Detects covered feet in the image.
           * - ARMPITS_COVERED
             - Detects covered armpits in the image.
           * - ARMPITS_EXPOSED
             - Detects exposed armpits in the image.
           * - FACE_MALE
             - Detects the face of a male in the image.
           * - BELLY_EXPOSED
             - Detects an exposed belly in the image.
           * - MALE_GENITALIA_EXPOSED
             - Detects exposed male genitalia in the image.
           * - ANUS_COVERED
             - Detects a covered anus in the image.
           * - FEMALE_BREAST_COVERED
             - Detects covered female breasts in the image.
           * - BUTTOCKS_COVERED
             - Detects covered buttocks in the image.


    .. note::
    
        This module requires onnxruntime version 1.18 or higher.
"""

from typing import Tuple, List

import numpy as np
from PIL import Image
from hbutils.testing.requires.version import VersionInfo
from huggingface_hub import hf_hub_download

from imgutils.data import ImageTyping
from imgutils.utils import open_onnx_model, ts_lru_cache
from ..data import load_image


def _check_compatibility() -> bool:
    """
    Check if the installed onnxruntime version is compatible with NudeNet.

    :raises EnvironmentError: If the onnxruntime version is less than 1.18.
    """
    import onnxruntime
    if VersionInfo(onnxruntime.__version__) < '1.18':
        raise EnvironmentError(f'Nudenet not supported on onnxruntime {onnxruntime.__version__}, '
                               f'please upgrade it to 1.18+ version.\n'
                               f'If you are running on CPU, use "pip install -U onnxruntime" .\n'
                               f'If you are running on GPU, use "pip install -U onnxruntime-gpu" .')  # pragma: no cover


_REPO_ID = 'deepghs/nudenet_onnx'


@ts_lru_cache()
def _open_nudenet_yolo():
    """
    Open and cache the NudeNet YOLO ONNX model.

    :return: The loaded ONNX model for YOLO.
    """
    return open_onnx_model(hf_hub_download(
        repo_id=_REPO_ID,
        repo_type='model',
        filename='320n.onnx',
    ))


@ts_lru_cache()
def _open_nudenet_nms():
    """
    Open and cache the NudeNet NMS ONNX model.

    :return: The loaded ONNX model for NMS.
    """
    return open_onnx_model(hf_hub_download(
        repo_id=_REPO_ID,
        repo_type='model',
        filename='nms-yolov8.onnx',
    ))


def _nn_preprocessing(image: ImageTyping, model_size: int = 320) -> Tuple[np.ndarray, float]:
    """
    Preprocess the input image for the NudeNet model.

    :param image: The input image.
    :param model_size: The size to which the image should be resized (default: 320).
    :return: A tuple containing the preprocessed image array and the scaling ratio.
    """
    image = load_image(image, mode='RGB', force_background='white')
    assert image.mode == 'RGB'
    mat = np.array(image)

    max_size = max(image.width, image.height)

    mat_pad = np.zeros((max_size, max_size, 3), dtype=np.uint8)
    mat_pad[:mat.shape[0], :mat.shape[1], :] = mat
    img_resized = Image.fromarray(mat_pad, mode='RGB').resize((model_size, model_size), resample=Image.BILINEAR)

    input_data = np.array(img_resized).transpose(2, 0, 1).astype(np.float32) / 255.0
    input_data = np.expand_dims(input_data, axis=0)
    return input_data, max_size / model_size


def _make_np_config(topk: int = 100, iou_threshold: float = 0.45, score_threshold: float = 0.25) -> np.ndarray:
    """
    Create a configuration array for the NMS model.

    :param topk: The maximum number of detections to keep (default: 100).
    :param iou_threshold: The IoU threshold for NMS (default: 0.45).
    :param score_threshold: The score threshold for detections (default: 0.25).
    :return: A numpy array containing the configuration parameters.
    """
    return np.array([topk, iou_threshold, score_threshold]).astype(np.float32)


def _nn_postprocess(selected, global_ratio: float):
    """
    Postprocess the model output to generate bounding boxes and labels.

    :param selected: The output from the NMS model.
    :param global_ratio: The scaling ratio to apply to the bounding boxes.
    :return: A list of tuples, each containing a bounding box, label, and confidence score.
    """
    bboxes = []
    num_boxes = selected.shape[0]
    for idx in range(num_boxes):
        data = selected[idx, :]

        scores = data[4:]
        score = np.max(scores)
        label = np.argmax(scores)

        box = data[:4] * global_ratio
        x = (box[0] - 0.5 * box[2]).item()
        y = (box[1] - 0.5 * box[3]).item()
        w = box[2].item()
        h = box[3].item()

        bboxes.append(((x, y, x + w, y + h), _LABELS[label], score.item()))

    return bboxes


_LABELS = [
    "FEMALE_GENITALIA_COVERED",
    "FACE_FEMALE",
    "BUTTOCKS_EXPOSED",
    "FEMALE_BREAST_EXPOSED",
    "FEMALE_GENITALIA_EXPOSED",
    "MALE_BREAST_EXPOSED",
    "ANUS_EXPOSED",
    "FEET_EXPOSED",
    "BELLY_COVERED",
    "FEET_COVERED",
    "ARMPITS_COVERED",
    "ARMPITS_EXPOSED",
    "FACE_MALE",
    "BELLY_EXPOSED",
    "MALE_GENITALIA_EXPOSED",
    "ANUS_COVERED",
    "FEMALE_BREAST_COVERED",
    "BUTTOCKS_COVERED"
]


def detect_with_nudenet(image: ImageTyping, topk: int = 100,
                        iou_threshold: float = 0.45, score_threshold: float = 0.25) \
        -> List[Tuple[Tuple[int, int, int, int], str, float]]:
    """
    Detect nudity in the given image using the NudeNet model.

    :param image: The input image to analyze.
    :param topk: The maximum number of detections to keep (default: 100).
    :param iou_threshold: The IoU threshold for NMS (default: 0.45).
    :param score_threshold: The score threshold for detections (default: 0.25).
    :return: A list of tuples, each containing:

             - A bounding box as (x1, y1, x2, y2)
             - A label string
             - A confidence score
    """
    _check_compatibility()
    input_, global_ratio = _nn_preprocessing(image, model_size=320)
    config = _make_np_config(topk, iou_threshold, score_threshold)
    output0, = _open_nudenet_yolo().run(['output0'], {'images': input_})
    selected, = _open_nudenet_nms().run(['selected'], {'detection': output0, 'config': config})
    return _nn_postprocess(selected[0], global_ratio=global_ratio)



"""
Overview:
    Metadata parser and formatter for a1111's webui generated images.

    Here are 2 sample images: :download:`sd_metadata_simple.png <sd_metadata_simple.png>`
    and :download:`sd_metadata_complex.png <sd_metadata_complex.png>`.
"""
import io
import json
import mimetypes
import os
import re
import textwrap
from dataclasses import dataclass
from typing import Dict, Any, Optional, Union

from PIL import Image
from PIL.PngImagePlugin import PngInfo

from ..data import ImageTyping, load_image
from ..metadata import read_geninfo_parameters, read_geninfo_exif, write_geninfo_exif, write_geninfo_gif, \
    read_geninfo_gif

_PARAM_PATTERN = re.compile(r'\s*(?P<key>[\w ]+):\s*(?P<value>"(?:\\.|[^\\"])+"|[^,]*)(?:,|$)')
_SIZE_PATTERN = re.compile(r"^(?P<size1>-?\d+)\s*x\s*(?P<size2>-?\d+)$")
_NEG_LINE_PATTERN = re.compile(r'^Negative prompt:\s*(?P<content>[\S\s]*?)$')

_PRE_KEYS = [
    "Steps", "Sampler", "CFG scale", "Image CFG scale", "Seed", "Face restoration", "Size",
    "Model hash", "Model", "VAE hash", "VAE", "Variation seed", "Variation seed strength",
    "Seed resize from", "Denoising strength", "Conditional mask weight", "Clip skip", "ENSD",
    "Token merging ratio", "Token merging ratio hr", "Init image hash", "RNG", "NGMS", "Tiling",
]
_POST_KEYS = [
    "Version", "User"
]


def _sdmeta_quote(value):
    if isinstance(value, tuple) and len(value) == 2:
        return f'{value[0]}x{value[1]}'
    if ',' not in str(value) and '\n' not in str(value) and ':' not in str(value):
        return value

    return json.dumps(value, ensure_ascii=False)


@dataclass
class SDMetaData:
    """
    Store information parsed from the metadata of a PNG image.

    :param prompt: The main prompt text.
    :type prompt: str
    :param neg_prompt: The negative prompt text.
    :type neg_prompt: str
    :param parameters: A dictionary containing various parameters.
    :type parameters: Dict[str, Any]
    """

    prompt: str
    neg_prompt: str
    parameters: Dict[str, Any]

    def __str__(self):
        """
        Get a string representation of the metadata.

        :return: A string representation of the metadata.
        :rtype: str

        Examples::
            >>> from imgutils.sd import get_sdmeta_from_image
            >>>
            >>> sd1 = get_sdmeta_from_image('sd_metadata_simple.png')
            >>> print(sd1)
            (extremely delicate and beautiful), best quality, official art, global illumination, soft shadow, super detailed, Japanese light novel cover, 4K, metal_texture, (striped_background), super detailed background, more detailed, rich detailed, extremely detailed CG unity 8k wallpaper, ((unreal)), sci-fi,(fantasy),(masterpiece),(super delicate), (illustration), (extremely delicate and beautiful), anime coloring,
            (silver_skin), ((high-cut silver_impossible_bodysuit), ((gem_on_chest)),(high-cut_silver_mechanical_leotard)),headgear,
            (focus-on:1.1),(1_girl),((solo)),slim_waist,white hair, long hair, luminous yellow eyes,(medium_breast:1.2), (Indistinct_cameltoe:0.9), (flat_crotch:1.1),(coquettish), (squint:1.4),(evil_smile :1.35),(dark_persona), [open mouth: 0.7], standing,[wet:0.7],
            slim_face, tall_girl,(mature),mature_face, (slim_figure), (slim_legs:1.1), (groin:1.1), ((bare_thighs)),
            Negative prompt: EasyNegative, sketch, duplicate, ugly, huge eyes, text, logo, monochrome, worst face, (bad and mutated hands:1.3), (worst quality:2.0), (low quality:2.0), (blurry:2.0), horror, geometry, bad_prompt, (bad hands), (missing fingers), multiple limbs, bad anatomy, (interlocked fingers:1.2), Ugly Fingers, (extra digit and hands and fingers and legs and arms:1.4), ((2girl)), (deformed fingers:1.2), (long fingers:1.2),(bad-artist-anime), bad-artist, bad hand, blush, (lipstick),skindentation, tie, ((big_breast)), (nipple), thighhighs, pubic_hair, pussy, black and white,(3d), ((realistic)),blurry,nipple slip, (nipple), blush, head_out_of_frame,curvy,
            Steps: 20, Sampler: DDIM, CFG scale: 7, Seed: 3827064803, Size: 512x848, Model hash: eb49192009, Model: AniDosMix, Clip skip: 2
        """
        return self._sdmeta_text()

    def _sdmeta_text(self):
        """
        Generate a formatted string representation of the metadata.

        This internal method is used by __str__ and other methods to create a consistent
        string representation of the metadata.

        :return: A formatted string containing the metadata.
        :rtype: str
        """
        with io.StringIO() as sio:
            print(self.prompt, file=sio)
            if self.neg_prompt:
                print(f'Negative prompt: {self.neg_prompt}', file=sio)

            _ext_keys = set(self.parameters.keys()) - set(_PRE_KEYS) - set(_POST_KEYS)
            _params = {}
            for key in _PRE_KEYS:
                if key in self.parameters:
                    _params[key] = self.parameters[key]
            for key in self.parameters.keys():
                if key in _ext_keys:
                    _params[key] = self.parameters[key]
            for key in _POST_KEYS:
                if key in self.parameters:
                    _params[key] = self.parameters[key]

            if _params:
                print(", ".join([
                    k if k == v else f'{k}: {_sdmeta_quote(v)}'
                    for k, v in _params.items() if v is not None
                ]), file=sio)

            return sio.getvalue().strip()

    @property
    def text(self) -> str:
        """
        Get the metadata as a formatted string.

        This property provides a convenient way to access the string representation
        of the metadata without calling _sdmeta_text() directly.

        :return: A formatted string containing the metadata.
        :rtype: str

        Example:
            >>> metadata = SDMetaData(
            ...     prompt="A starry night",
            ...     neg_prompt="Daylight",
            ...     parameters={"Steps": 40, "Sampler": "Euler", "CFG scale": 8}
            ... )
            >>> print(metadata.text)
            A starry night
            Negative prompt: Daylight
            Steps: 40, Sampler: Euler, CFG scale: 8
        """
        return self._sdmeta_text()

    @property
    def pnginfo(self) -> PngInfo:
        """
        Generate a PngInfo object with the metadata.

        This can be used when saving an image with custom metadata.

        :return: A PngInfo object containing the metadata.
        :rtype: PngInfo

        Examples::
            >>> from PIL import Image
            >>> from imgutils.sd import get_sdmeta_from_image
            >>>
            >>> # get metadata
            >>> sd1 = get_sdmeta_from_image('sd_metadata_simple.png')
            >>>
            >>> # create an image
            >>> img = Image.new('RGB', (256, 256), 'white')
            >>>
            >>> # save this image with sd1's metadata
            >>> img.save('new_image.png', pnginfo=sd1.pnginfo)
            >>>
            >>> # let's see what is in 'new_image.png'
            >>> get_sdmeta_from_image('new_image.png')
            SDMetaData(prompt='(extremely delicate and beautiful), best quality, official art, global illumination, soft shadow, super detailed, Japanese light novel cover, 4K, metal_texture, (striped_background), super detailed background, more detailed, rich detailed, extremely detailed CG unity 8k wallpaper, ((unreal)), sci-fi,(fantasy),(masterpiece),(super delicate), (illustration), (extremely delicate and beautiful), anime coloring,\\n(silver_skin), ((high-cut silver_impossible_bodysuit), ((gem_on_chest)),(high-cut_silver_mechanical_leotard)),headgear,\\n(focus-on:1.1),(1_girl),((solo)),slim_waist,white hair, long hair, luminous yellow eyes,(medium_breast:1.2), (Indistinct_cameltoe:0.9), (flat_crotch:1.1),(coquettish), (squint:1.4),(evil_smile :1.35),(dark_persona), [open mouth: 0.7], standing,[wet:0.7],\\nslim_face, tall_girl,(mature),mature_face, (slim_figure), (slim_legs:1.1), (groin:1.1), ((bare_thighs)),', neg_prompt='EasyNegative, sketch, duplicate, ugly, huge eyes, text, logo, monochrome, worst face, (bad and mutated hands:1.3), (worst quality:2.0), (low quality:2.0), (blurry:2.0), horror, geometry, bad_prompt, (bad hands), (missing fingers), multiple limbs, bad anatomy, (interlocked fingers:1.2), Ugly Fingers, (extra digit and hands and fingers and legs and arms:1.4), ((2girl)), (deformed fingers:1.2), (long fingers:1.2),(bad-artist-anime), bad-artist, bad hand, blush, (lipstick),skindentation, tie, ((big_breast)), (nipple), thighhighs, pubic_hair, pussy, black and white,(3d), ((realistic)),blurry,nipple slip, (nipple), blush, head_out_of_frame,curvy,', parameters={'Steps': 20, 'Sampler': 'DDIM', 'CFG scale': 7, 'Seed': 3827064803, 'Size': (512, 848), 'Model hash': 'eb49192009', 'Model': 'AniDosMix', 'Clip skip': 2})
        """
        info = PngInfo()
        info.add_text('parameters', self._sdmeta_text())
        return info


def _parse_parameters(param_text: str):
    params = {}
    for key, value in _PARAM_PATTERN.findall(param_text):
        try:
            value = json.loads(value)
        except json.JSONDecodeError:
            pass

        if isinstance(value, str):
            size_matching = _SIZE_PATTERN.match(value)
            if size_matching is not None:
                params[key] = (
                    int(size_matching.group('size1')),
                    int(size_matching.group('size2'))
                )
            else:
                params[key] = value
        else:
            params[key] = value

    return params


def parse_sdmeta_from_text(x: str) -> SDMetaData:
    """
    Parse metadata information from a string.

    :param x: The input string containing metadata.
    :type x: str
    :return: A SDMetaData object containing the parsed metadata.
    :rtype: SDMetaData

    Examples::
        >>> from imgutils.sd import parse_sdmeta_from_text
        >>>
        >>> sd1 = parse_sdmeta_from_text(\"\"\"
        ... (extremely delicate and beautiful), best quality, official art, global illumination, soft shadow, super detailed, Japanese light novel cover, 4K, metal_texture, (striped_background), super detailed background, more detailed, rich detailed, extremely detailed CG unity 8k wallpaper, ((unreal)), sci-fi,(fantasy),(masterpiece),(super delicate), (illustration), (extremely delicate and beautiful), anime coloring,
        ... (silver_skin), ((high-cut silver_impossible_bodysuit), ((gem_on_chest)),(high-cut_silver_mechanical_leotard)),headgear,
        ... (focus-on:1.1),(1_girl),((solo)),slim_waist,white hair, long hair, luminous yellow eyes,(medium_breast:1.2), (Indistinct_cameltoe:0.9), (flat_crotch:1.1),(coquettish), (squint:1.4),(evil_smile :1.35),(dark_persona), [open mouth: 0.7], standing,[wet:0.7],
        ... slim_face, tall_girl,(mature),mature_face, (slim_figure), (slim_legs:1.1), (groin:1.1), ((bare_thighs)),
        ... Negative prompt: EasyNegative, sketch, duplicate, ugly, huge eyes, text, logo, monochrome, worst face, (bad and mutated hands:1.3), (worst quality:2.0), (low quality:2.0), (blurry:2.0), horror, geometry, bad_prompt, (bad hands), (missing fingers), multiple limbs, bad anatomy, (interlocked fingers:1.2), Ugly Fingers, (extra digit and hands and fingers and legs and arms:1.4), ((2girl)), (deformed fingers:1.2), (long fingers:1.2),(bad-artist-anime), bad-artist, bad hand, blush, (lipstick),skindentation, tie, ((big_breast)), (nipple), thighhighs, pubic_hair, pussy, black and white,(3d), ((realistic)),blurry,nipple slip, (nipple), blush, head_out_of_frame,curvy,
        ... Steps: 20, Sampler: DDIM, CFG scale: 7, Seed: 3827064803, Size: 512x848, Model hash: eb49192009, Model: AniDosMix, Clip skip: 2
        ... \"\"\")
        >>> sd1
        SDMetaData(prompt='(extremely delicate and beautiful), best quality, official art, global illumination, soft shadow, super detailed, Japanese light novel cover, 4K, metal_texture, (striped_background), super detailed background, more detailed, rich detailed, extremely detailed CG unity 8k wallpaper, ((unreal)), sci-fi,(fantasy),(masterpiece),(super delicate), (illustration), (extremely delicate and beautiful), anime coloring,\\n(silver_skin), ((high-cut silver_impossible_bodysuit), ((gem_on_chest)),(high-cut_silver_mechanical_leotard)),headgear,\\n(focus-on:1.1),(1_girl),((solo)),slim_waist,white hair, long hair, luminous yellow eyes,(medium_breast:1.2), (Indistinct_cameltoe:0.9), (flat_crotch:1.1),(coquettish), (squint:1.4),(evil_smile :1.35),(dark_persona), [open mouth: 0.7], standing,[wet:0.7],\\nslim_face, tall_girl,(mature),mature_face, (slim_figure), (slim_legs:1.1), (groin:1.1), ((bare_thighs)),', neg_prompt='EasyNegative, sketch, duplicate, ugly, huge eyes, text, logo, monochrome, worst face, (bad and mutated hands:1.3), (worst quality:2.0), (low quality:2.0), (blurry:2.0), horror, geometry, bad_prompt, (bad hands), (missing fingers), multiple limbs, bad anatomy, (interlocked fingers:1.2), Ugly Fingers, (extra digit and hands and fingers and legs and arms:1.4), ((2girl)), (deformed fingers:1.2), (long fingers:1.2),(bad-artist-anime), bad-artist, bad hand, blush, (lipstick),skindentation, tie, ((big_breast)), (nipple), thighhighs, pubic_hair, pussy, black and white,(3d), ((realistic)),blurry,nipple slip, (nipple), blush, head_out_of_frame,curvy,', parameters={'Steps': 20, 'Sampler': 'DDIM', 'CFG scale': 7, 'Seed': 3827064803, 'Size': (512, 848), 'Model hash': 'eb49192009', 'Model': 'AniDosMix', 'Clip skip': 2})
        >>> type(sd1)
        <class 'imgutils.sd.metadata.SDMetaData'>
        >>>
        >>> sd2 = parse_sdmeta_from_text(\"\"\"
        ... 1girl, solo, blue eyes, black footwear, white hair, looking at viewer, shoes, full body, standing, bangs, indoors, wide sleeves, ahoge, dress, closed mouth, blush, long sleeves, potted plant, bag, plant, hair bun, window,<lora:BlueArchive10:1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1>,BlueArchive,
        ... Negative prompt: Neg1,Negative,
        ... Steps: 20, Sampler: DPM++ 2M SDE Karras, CFG scale: 7, Seed: 2647703743, Size: 768x768, Model hash: 72bd94132e, Model: CuteMix, Denoising strength: 0.7, ControlNet 0: "preprocessor: openpose, model: control_v11p_sd15_openpose [cab727d4], weight: 1, starting/ending: (0, 1), resize mode: Crop and Resize, pixel perfect: False, control mode: Balanced, preprocessor params: (512, 64, 64)", Hires upscale: 2, Hires upscaler: Latent, TI hashes: "Neg1: 339cc9210f70, Negative: 66a7279a88dd", Version: v1.5.1
        ... \"\"\")
        >>> sd2
        SDMetaData(prompt='1girl, solo, blue eyes, black footwear, white hair, looking at viewer, shoes, full body, standing, bangs, indoors, wide sleeves, ahoge, dress, closed mouth, blush, long sleeves, potted plant, bag, plant, hair bun, window,<lora:BlueArchive10:1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1>,BlueArchive,', neg_prompt='Neg1,Negative,', parameters={'Steps': 20, 'Sampler': 'DPM++ 2M SDE Karras', 'CFG scale': 7, 'Seed': 2647703743, 'Size': (768, 768), 'Model hash': '72bd94132e', 'Model': 'CuteMix', 'Denoising strength': 0.7, 'ControlNet 0': 'preprocessor: openpose, model: control_v11p_sd15_openpose [cab727d4], weight: 1, starting/ending: (0, 1), resize mode: Crop and Resize, pixel perfect: False, control mode: Balanced, preprocessor params: (512, 64, 64)', 'Hires upscale': 2, 'Hires upscaler': 'Latent', 'TI hashes': 'Neg1: 339cc9210f70, Negative: 66a7279a88dd', 'Version': 'v1.5.1'})
        >>> type(sd2)
        <class 'imgutils.sd.metadata.SDMetaData'>
    """
    x = textwrap.dedent(x).strip()
    all_lines = x.splitlines(keepends=False)
    if len(all_lines) >= 1:
        prompt_lines = all_lines[:-1]
        argument_line = all_lines[-1]
    else:
        prompt_lines = []
        argument_line = ''
    if len(_PARAM_PATTERN.findall(argument_line)) < 3:
        prompt_lines.append(argument_line)
        argument_line = ''

    # 0x1 means prompt, 0x2 means neg prompt
    status = 0x1
    with io.StringIO() as i_prompt, io.StringIO() as i_neg_prompt:
        for line in prompt_lines:
            line = line.strip()
            neg_matching = _NEG_LINE_PATTERN.fullmatch(line)
            if neg_matching:
                line = neg_matching.group('content').strip()
                status = 0x2

            if status == 0x1:
                print(line, file=i_prompt)
            elif status == 0x2:
                print(line, file=i_neg_prompt)
            else:
                # if this line is triggered, this state machine must be buggy
                raise ValueError(f'Unknown parsing status - {status!r}.')  # pragma: no cover

        prompt = i_prompt.getvalue().strip()
        neg_prompt = i_neg_prompt.getvalue().strip()

    params = _parse_parameters(argument_line)
    return SDMetaData(prompt, neg_prompt, params)


class _InvalidSDMetaError(Exception):
    """
    Custom exception raised when SD metadata is invalid or not found.

    This exception is used internally to signal that the metadata
    validation or extraction process has failed.
    """
    pass


def _sdtext_validate(text: str) -> str:
    """
    Validate the given text as SD metadata.

    This function attempts to validate the input text as SD metadata. It first tries
    to validate it as NAI (Novel AI) metadata, and if that fails, it checks if the
    text is non-empty. If both checks fail, it raises an _InvalidSDMetaError.

    :param text: The text to validate as SD metadata.
    :type text: str

    :return: The validated text if it passes the checks.
    :rtype: str

    :raises _InvalidSDMetaError: If the text is invalid or empty.
    """
    from .nai import _naimeta_text_validate, _InvalidNAIMetaError

    try:
        _naimeta_text_validate(text)
    except _InvalidNAIMetaError:
        pass
    else:
        raise _InvalidSDMetaError

    if text:
        return text
    else:
        raise _InvalidSDMetaError


def _get_raw_sdtext(image: ImageTyping) -> Optional[str]:
    """
    Extract raw SD metadata text from the given image.

    This function attempts to read SD metadata from various sources within the image,
    including PNG info, EXIF data, and GIF metadata. It tries each method in turn
    and returns the first valid SD metadata text found.

    :param image: The input image to extract metadata from.
    :type image: ImageTyping

    :return: The raw SD metadata text if found, otherwise None.
    :rtype: Optional[str]

    :raises _InvalidSDMetaError: If no valid SD metadata is found in the image.
    """
    image = load_image(image, force_background=None, mode=None)

    try:
        return _sdtext_validate(read_geninfo_parameters(image))
    except _InvalidSDMetaError:
        pass

    try:
        return _sdtext_validate(read_geninfo_exif(image))
    except _InvalidSDMetaError:
        pass

    try:
        return _sdtext_validate(read_geninfo_gif(image))
    except _InvalidSDMetaError:
        raise _InvalidSDMetaError


def get_sdmeta_from_image(image: ImageTyping) -> Optional[SDMetaData]:
    """
    Extract and parse Stable Diffusion metadata from an image.

    This function attempts to read SD metadata from various sources within the image,
    including PNG info, EXIF data, and GIF metadata. If found, it parses the metadata
    into an SDMetaData object.

    :param image: The input image, which can be a file path, URL, or PIL Image object.
    :type image: ImageTyping

    :return: An SDMetaData object containing the parsed metadata if available, else None.
    :rtype: Optional[SDMetaData]

    :raises: Various exceptions may be raised by the underlying image loading and
             metadata reading functions.

    Example usage:
        >>> from imgutils.sd import get_sdmeta_from_image
        >>> sd_meta = get_sdmeta_from_image('path/to/image.png')
        >>> if sd_meta:
        ...     print(f"Prompt: {sd_meta.prompt}")
        ...     print(f"Negative prompt: {sd_meta.neg_prompt}")
        ...     print(f"Parameters: {sd_meta.parameters}")
        ... else:
        ...     print("No SD metadata found in the image.")

    Note: This function depends on the load_image and parse_sdmeta_from_text functions.
    Ensure these are properly imported or defined in the current scope.
    """
    image = load_image(image, mode=None, force_background=None)
    try:
        pnginfo_text = _get_raw_sdtext(image)
    except _InvalidSDMetaError:
        return None
    else:
        return parse_sdmeta_from_text(pnginfo_text)


def _save_png_with_sdmeta(image: Image.Image, dst_file: Union[str, os.PathLike], metadata: SDMetaData, **kwargs):
    """
    Internal function to save a PNG image with SD metadata.

    :param image: The PIL Image object to save.
    :param dst_file: The destination file path.
    :param metadata: The SDMetaData object containing the metadata to save.
    :param kwargs: Additional keyword arguments to pass to the PIL save function.
    """
    image.save(dst_file, pnginfo=metadata.pnginfo, **kwargs)


def _save_exif_with_sdmeta(image: Image.Image, dst_file: Union[str, os.PathLike], metadata: SDMetaData, **kwargs):
    """
    Internal function to save an image with SD metadata in EXIF format.

    :param image: The PIL Image object to save.
    :param dst_file: The destination file path.
    :param metadata: The SDMetaData object containing the metadata to save.
    :param kwargs: Additional keyword arguments to pass to the write_geninfo_exif function.
    """
    write_geninfo_exif(image, dst_file, metadata.text, **kwargs)


def _save_gif_with_sdmeta(image: Image.Image, dst_file: Union[str, os.PathLike], metadata: SDMetaData, **kwargs):
    """
    Internal function to save a GIF image with SD metadata.

    :param image: The PIL Image object to save.
    :param dst_file: The destination file path.
    :param metadata: The SDMetaData object containing the metadata to save.
    :param kwargs: Additional keyword arguments to pass to the write_geninfo_gif function.
    """
    write_geninfo_gif(image, dst_file, metadata.text, **kwargs)


_FN_IMG_SAVE = {
    'image/png': _save_png_with_sdmeta,
    'image/jpeg': _save_exif_with_sdmeta,
    'image/webp': _save_exif_with_sdmeta,
    'image/gif': _save_gif_with_sdmeta,
}


def save_image_with_sdmeta(image: ImageTyping, dst_file: Union[str, os.PathLike], metadata: SDMetaData, **kwargs):
    """
    Save an image with Stable Diffusion metadata.

    This function saves the given image to the specified destination file, including
    the provided SD metadata. The metadata is saved in a format appropriate for the
    output image type (PNG, JPEG, WebP, or GIF).

    :param image: The input image, which can be a file path, URL, or PIL Image object.
    :type image: ImageTyping
    :param dst_file: The destination file path where the image will be saved.
    :type dst_file: Union[str, os.PathLike]
    :param metadata: The SD metadata to include with the image.
    :type metadata: SDMetaData
    :param kwargs: Additional keyword arguments to pass to the underlying save function.

    :raises SystemError: If the output file type is not supported for saving with metadata.
    :raises: Various exceptions may be raised by the underlying image loading and
             saving functions.

    Example usage:
        >>> from imgutils.sd import get_sdmeta_from_image, save_image_with_sdmeta
        >>> input_image = 'path/to/input.png'
        >>> output_image = 'path/to/output.png'
        >>> sd_meta = get_sdmeta_from_image(input_image)
        >>> if sd_meta:
        ...     save_image_with_sdmeta(input_image, output_image, sd_meta)
        ...     print(f"Image saved with SD metadata to {output_image}")
        ... else:
        ...     print("No SD metadata found in the input image.")
    """
    mimetype, _ = mimetypes.guess_type(str(dst_file))
    if mimetype not in _FN_IMG_SAVE:
        raise SystemError(f'Not supported to save as a {mimetype!r} type, '
                          f'supported mimetypes are {sorted(_FN_IMG_SAVE.keys())!r}.')

    image = load_image(image, mode=None, force_background=None)
    _FN_IMG_SAVE[mimetype](image, dst_file, metadata, **kwargs)



"""
Overview:
    Detect human censor points (including female's nipples and genitals of both male and female) in anime images.

    Trained on dataset `deepghs/anime_censor_detection <https://huggingface.co/datasets/deepghs/anime_censor_detection>`_ with YOLOv8.

    .. collapse:: Overview of Censor Detect (NSFW Warning!!!)

        .. image:: censor_detect_demo.plot.py.svg
            :align: center

    This is an overall benchmark of all the censor detect models:

    .. image:: censor_detect_benchmark.plot.py.svg
        :align: center

"""
from typing import List, Tuple, Optional

from ..data import ImageTyping
from ..generic import yolo_predict

_REPO_ID = 'deepghs/anime_censor_detection'


def detect_censors(image: ImageTyping, level: str = 's', version: str = 'v1.0', model_name: Optional[str] = None,
                   conf_threshold: float = 0.3, iou_threshold: float = 0.7, **kwargs) \
        -> List[Tuple[Tuple[int, int, int, int], str, float]]:
    """
    Detect human censor points in anime images.

    This function uses pre-trained YOLOv8 models to identify and locate specific
    anatomical features that are typically censored in anime images. It can detect
    female nipples, male genitals, and female genitals.

    :param image: The input image to be analyzed. Can be a file path, URL, or image data.
    :type image: ImageTyping

    :param level: The model level to use, either 's' (standard) or 'n' (nano).
                  The 'n' model is faster but less accurate, while 's' is more accurate but slower.
    :type level: str

    :param version: The version of the model to use. Default is 'v1.0'.
    :type version: str

    :param model_name: Optional custom model name. If not provided, it will be constructed
                       from the version and level.
    :type model_name: Optional[str]

    :param conf_threshold: The confidence threshold for detections. Only detections with
                           confidence above this value will be returned. Default is 0.3.
    :type conf_threshold: float

    :param iou_threshold: The Intersection over Union (IoU) threshold for non-maximum
                          suppression. Detections with IoU above this value will be merged.
                          Default is 0.7.
    :type iou_threshold: float

    :return: A list of tuples, each containing:
             - A tuple of four integers (x0, y0, x1, y1) representing the bounding box
             - A string indicating the type of detection ('nipple_f', 'penis', or 'pussy')
             - A float representing the confidence score of the detection
    :rtype: List[Tuple[Tuple[int, int, int, int], str, float]]

    :raises ValueError: If an invalid level is provided.
    :raises RuntimeError: If the model fails to load or process the image.

    Examples::
        >>> from imgutils.detect import detect_censors, detection_visualize
        >>>
        >>> image = 'nude_girl.png'
        >>> result = detect_censors(image)  # detect it
        >>> result
        [
            ((365, 264, 399, 289), 'nipple_f', 0.7473511695861816),
            ((224, 260, 252, 285), 'nipple_f', 0.6830288171768188),
            ((206, 523, 240, 608), 'pussy', 0.6799028515815735)
        ]
        >>>
        >>> # visualize it
        >>> from matplotlib import pyplot as plt
        >>> plt.imshow(detection_visualize(image, result))
        >>> plt.show()
    """
    return yolo_predict(
        image=image,
        repo_id=_REPO_ID,
        model_name=model_name or f'censor_detect_{version}_{level}',
        conf_threshold=conf_threshold,
        iou_threshold=iou_threshold,
        **kwargs,
    )



"""
Overview:
    Obtaining the outline (or you can call that line drawing) of an anime image.

    Here is the example and comparison:

    .. image:: edge_demo.plot.py.svg
        :align: center

    This is an overall benchmark of all the outline models:

    .. image:: edge_benchmark.plot.py.svg
        :align: center

"""
from .canny import get_edge_by_canny, edge_image_with_canny
from .lineart import get_edge_by_lineart, edge_image_with_lineart
from .lineart_anime import get_edge_by_lineart_anime, edge_image_with_lineart_anime



"""
Overview:
    A tool for measuring the aesthetic level of anime images, with the model
    obtained from `skytnt/anime-aesthetic <https://huggingface.co/skytnt/anime-aesthetic>`_.

    .. image:: aesthetic_full.plot.py.svg
        :align: center

    This is an overall benchmark of all the operations in aesthetic models:

    .. image:: aesthetic_benchmark.plot.py.svg
        :align: center

    .. warning::
        These model is deprecated due to the poor effectiveness.
        Please use `imgutils.metrics.aesthetic.anime_dbaesthetic` for better evaluation.
"""

import cv2
import numpy as np
from PIL import Image
from deprecation import deprecated
from huggingface_hub import hf_hub_download

from ..config.meta import __VERSION__
from ..data import ImageTyping, load_image
from ..utils import open_onnx_model, ts_lru_cache

__all__ = [
    'get_aesthetic_score',
]


@ts_lru_cache()
def _open_aesthetic_model():
    return open_onnx_model(hf_hub_download(
        repo_id="skytnt/anime-aesthetic",
        filename="model.onnx"
    ))


def _preprocess(image: Image.Image):
    assert image.mode == 'RGB'
    img = np.array(image).astype(np.float32) / 255
    s = 768
    h, w = img.shape[:-1]
    h, w = (s, int(s * w / h)) if h > w else (int(s * h / w), s)
    ph, pw = s - h, s - w
    img_input = np.zeros([s, s, 3], dtype=np.float32)
    img_input[ph // 2:ph // 2 + h, pw // 2:pw // 2 + w] = cv2.resize(img, (w, h))
    img_input = np.transpose(img_input, (2, 0, 1))
    return img_input[np.newaxis, :]


@deprecated(deprecated_in='0.4.2', removed_in='1.0.0', current_version=__VERSION__,
            details='Deprecated due to the low effectiveness.')
def get_aesthetic_score(image: ImageTyping):
    """
    Overview:
        Get aesthetic score for image.

    :param image: Original image.
    :return: Score of aesthetic.

    Examples::
        >>> from imgutils.metrics import get_aesthetic_score
        >>>
        >>> get_aesthetic_score('2053756.jpg')
        0.09986039996147156
        >>> get_aesthetic_score('1663584.jpg')
        0.24299287796020508
        >>> get_aesthetic_score('4886411.jpg')
        0.38091593980789185
        >>> get_aesthetic_score('2066024.jpg')
        0.5131649971008301
        >>> get_aesthetic_score('3670169.jpg')
        0.6011670827865601
        >>> get_aesthetic_score('5930006.jpg')
        0.7067991495132446
        >>> get_aesthetic_score('3821265.jpg')
        0.8237218260765076
        >>> get_aesthetic_score('5512471.jpg')
        0.9187621474266052
    """
    image = load_image(image, mode='RGB')
    retval, *_ = _open_aesthetic_model().run(None, {'img': _preprocess(image)})
    return float(retval.item())



"""
Overview:
    Detect eyes in anime images.

    Trained on dataset `deepghs/anime_eye_detection <https://huggingface.co/datasets/deepghs/anime_eye_detection>`_ with YOLOv8.

    .. image:: eye_detect_demo.plot.py.svg
        :align: center

    This is an overall benchmark of all the eye detect models:

    .. image:: eye_detect_benchmark.plot.py.svg
        :align: center

"""
from typing import List, Tuple, Optional

from ..data import ImageTyping
from ..generic import yolo_predict

_REPO_ID = 'deepghs/anime_eye_detection'


def detect_eyes(image: ImageTyping, level: str = 's', version: str = 'v1.0', model_name: Optional[str] = None,
                conf_threshold: float = 0.3, iou_threshold: float = 0.3, **kwargs) \
        -> List[Tuple[Tuple[int, int, int, int], str, float]]:
    """
    Detect human eyes in anime images.

    This function uses a YOLOv8 model to detect eyes in the given anime image. It supports
    different model levels and versions, allowing for a trade-off between speed and accuracy.

    :param image: The input image for eye detection. Can be various image types supported by ImageTyping.
    :type image: ImageTyping

    :param level: The model level to use. Can be either 's' (for higher accuracy) or 'n' (for faster processing).
                  Default is 's'.
    :type level: str

    :param version: Version of the model to use. Default is 'v1.0'.
    :type version: str

    :param model_name: Optional custom model name. If not provided, it's constructed using version and level.
    :type model_name: Optional[str]

    :param conf_threshold: Confidence threshold for detections. Only detections with confidence above this
                           threshold are returned. Default is 0.3.
    :type conf_threshold: float

    :param iou_threshold: Intersection over Union (IoU) threshold for non-maximum suppression.
                          Detections with IoU above this threshold are considered overlapping and merged.
                          Default is 0.3.
    :type iou_threshold: float

    :return: A list of detected eyes. Each detection is represented by a tuple containing:
             - Bounding box coordinates as (x0, y0, x1, y1)
             - Detection class (always 'eye' for this function)
             - Confidence score of the detection
    :rtype: List[Tuple[Tuple[int, int, int, int], str, float]]

    :raises: May raise exceptions related to image loading or model prediction (from yolo_predict function).

    Examples::
        >>> from imgutils.detect import detect_eyes, detection_visualize
        >>>
        >>> image = 'squat.jpg'
        >>> result = detect_eyes(image)  # detect it
        >>> result
        [((297, 239, 341, 271), 'eye', 0.7760562896728516), ((230, 289, 263, 308), 'eye', 0.7682342529296875)]
        >>>
        >>> # visualize it
        >>> from matplotlib import pyplot as plt
        >>> plt.imshow(detection_visualize(image, result))
        >>> plt.show()
    """
    return yolo_predict(
        image=image,
        repo_id=_REPO_ID,
        model_name=model_name or f'eye_detect_{version}_{level}',
        conf_threshold=conf_threshold,
        iou_threshold=iou_threshold,
        **kwargs,
    )



"""
Overview:
    A model for detecting AI-created images.

    The following are sample images for testing.

    .. image:: aicheck.plot.py.svg
        :align: center

    This is an overall benchmark of all the AI-check validation models:

    .. image:: aicheck_benchmark.plot.py.svg
        :align: center

    The models are hosted on
    `huggingface - deepghs/anime_ai_check <https://huggingface.co/deepghs/anime_ai_check>`_.
"""
from ..data import ImageTyping
from ..generic import classify_predict, classify_predict_score

__all__ = [
    'get_ai_created_score',
    'is_ai_created',
]

_DEFAULT_MODEL_NAME = 'mobilenetv3_sce_dist'
_REPO_ID = 'deepghs/anime_ai_check'


def get_ai_created_score(image: ImageTyping, model_name: str = _DEFAULT_MODEL_NAME) -> float:
    """
    Overview:
        Predict if the given image is created by AI (mainly by stable diffusion), given a score.

    :param image: Image to be predicted.
    :param model_name: Name of the model. Default is ``mobilenetv3_sce_dist``.
        If you need better accuracy, use ``caformer_s36_plus_sce``.
        All the available values are listed on the benchmark graph.
    :return: A float number which represent the score of AI-check.

    Examples::
        >>> from imgutils.validate import get_ai_created_score
        >>>
        >>> get_ai_created_score('aicheck/ai/1.jpg')
        0.9996960163116455
        >>> get_ai_created_score('aicheck/ai/2.jpg')
        0.9999125003814697
        >>> get_ai_created_score('aicheck/ai/3.jpg')
        0.997803270816803
        >>> get_ai_created_score('aicheck/ai/4.jpg')
        0.9960069060325623
        >>> get_ai_created_score('aicheck/ai/5.jpg')
        0.9887709021568298
        >>> get_ai_created_score('aicheck/ai/6.jpg')
        0.9998629093170166
        >>> get_ai_created_score('aicheck/human/7.jpg')
        0.0013722758740186691
        >>> get_ai_created_score('aicheck/human/8.jpg')
        0.00020673229300882667
        >>> get_ai_created_score('aicheck/human/9.jpg')
        0.0001895089662866667
        >>> get_ai_created_score('aicheck/human/10.jpg')
        0.0008857478387653828
        >>> get_ai_created_score('aicheck/human/11.jpg')
        4.552320024231449e-05
        >>> get_ai_created_score('aicheck/human/12.jpg')
        0.001168627175502479
    """
    return classify_predict_score(image, _REPO_ID, model_name)['ai']


def is_ai_created(image: ImageTyping, model_name: str = _DEFAULT_MODEL_NAME, threshold: float = 0.5) -> bool:
    """
    Overview:
        Predict if the given image is created by AI (mainly by stable diffusion).

    :param image: Image to be predicted.
    :param model_name: Name of the model. Default is ``mobilenetv3_sce_dist``.
        If you need better accuracy, use ``caformer_s36_plus_sce``.
        All the available values are listed on the benchmark graph.
    :param threshold: Threshold of the score. When the score is no less than ``threshold``, this image
        will be predicted as ``AI-created``. Default is ``0.5``.
    :return: This image is ``AI-created`` or not.

    Examples::
        >>> from imgutils.validate import is_ai_created
        >>>
        >>> is_ai_created('aicheck/ai/1.jpg')
        True
        >>> is_ai_created('aicheck/ai/2.jpg')
        True
        >>> is_ai_created('aicheck/ai/3.jpg')
        True
        >>> is_ai_created('aicheck/ai/4.jpg')
        True
        >>> is_ai_created('aicheck/ai/5.jpg')
        True
        >>> is_ai_created('aicheck/ai/6.jpg')
        True
        >>> is_ai_created('aicheck/human/7.jpg')
        False
        >>> is_ai_created('aicheck/human/8.jpg')
        False
        >>> is_ai_created('aicheck/human/9.jpg')
        False
        >>> is_ai_created('aicheck/human/10.jpg')
        False
        >>> is_ai_created('aicheck/human/11.jpg')
        False
        >>> is_ai_created('aicheck/human/12.jpg')
        False
    """
    type_, _ = classify_predict(image, _REPO_ID, model_name)
    return type_ == 'ai'



"""
Overview:
    Detect text in images.

    Models are hosted on `deepghs/text_detection <https://huggingface.co/spaces/deepghs/text_detection>`_.

    .. image:: text_detect_demo.plot.py.svg
        :align: center

    This is an overall benchmark of all the text detect models:

    .. image:: text_detect_benchmark.plot.py.svg
        :align: center

    .. warning::
        This module has been deprecated and will be removed in the future.

        It is recommended to migrate to the :func:`imgutils.ocr.detect_text_with_ocr` function as soon as possible.
        This function uses a higher-quality text detection model provided by PaddleOCR,
        resulting in improved performance and higher efficiency.

        .. image:: text_detect_deprecate_demo.plot.py.svg
            :align: center

"""
from typing import List, Tuple, Optional

import cv2
import numpy as np
from deprecation import deprecated
from huggingface_hub import hf_hub_download

from ..config.meta import __VERSION__
from ..data import ImageTyping, load_image
from ..utils import open_onnx_model, ts_lru_cache

_DEFAULT_MODEL = 'dbnetpp_resnet50_fpnc_1200e_icdar2015'


@ts_lru_cache()
def _open_text_detect_model(model: str):
    """
    Get an ONNX session for the specified DBNET or DBNET++ model.

    This function downloads the ONNX model and opens it using the imgutils library.

    :param model: Model name for DBNET or DBNET++.
    :type model: str
    :return: ONNX session for the specified model.
    """
    return open_onnx_model(hf_hub_download(
        'deepghs/text_detection',
        f'{model}/end2end.onnx'
    ))


def _get_heatmap_of_text(image: ImageTyping, model: str) -> np.ndarray:
    """
    Get the heatmap of text regions from the given image using the specified model.

    :param image: Input image.
    :type image: ImageTyping
    :param model: Model name for DBNET or DBNET++.
    :type model: str
    :return: Heatmap of text regions.
    :rtype: np.ndarray
    """
    origin_width, origin_height = width, height = image.size
    align = 32
    if width % align != 0:
        width += (align - width % align)
    if height % align != 0:
        height += (align - height % align)

    input_ = np.array(image).transpose((2, 0, 1)).astype(np.float32) / 255.0
    input_ = np.pad(input_[None, ...], ((0, 0), (0, 0), (0, height - origin_height), (0, width - origin_width)))

    def _normalize(data, mean=(0.48145466, 0.4578275, 0.40821073), std=(0.26862954, 0.26130258, 0.27577711)):
        mean, std = np.asarray(mean), np.asarray(std)
        return (data - mean[None, :, None, None]) / std[None, :, None, None]

    ort = _open_text_detect_model(model)

    input_ = _normalize(input_).astype(np.float32)
    output_, = ort.run(['output'], {'input': input_})
    heatmap = output_[0]
    heatmap = heatmap[:origin_height, :origin_width]

    return heatmap


def _get_bounding_box_of_text(image: ImageTyping, model: str, threshold: float) \
        -> List[Tuple[Tuple[int, int, int, int], float]]:
    """
    Get bounding boxes of detected text regions from the given image using the specified model and threshold.

    :param image: Input image.
    :type image: ImageTyping
    :param model: Model name for DBNET or DBNET++.
    :type model: str
    :param threshold: Confidence threshold for text detection.
    :type threshold: float
    :return: List of bounding boxes and their scores.
    :rtype: List[Tuple[Tuple[int, int, int, int], float]]
    """
    heatmap = _get_heatmap_of_text(image, model)
    c_rets = cv2.findContours((heatmap * 255.0).astype(np.uint8), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    contours = c_rets[0] if len(c_rets) == 2 else c_rets[1]
    bboxes = []
    for c in contours:
        x, y, w, h = cv2.boundingRect(c)
        x0, y0, x1, y1 = x, y, x + w, y + h
        score = heatmap[y0:y1, x0:x1].mean().item()
        if score >= threshold:
            bboxes.append(((x0, y0, x1, y1), score))

    return bboxes


@deprecated(deprecated_in="0.2.10", current_version=__VERSION__,
            details="Use the new function :func:`imgutils.ocr.detect_text_with_ocr` instead")
def detect_text(image: ImageTyping, model: str = _DEFAULT_MODEL, threshold: float = 0.05,
                max_area_size: Optional[int] = 640):
    """
    Detect text regions in the given image using the specified model and threshold.

    :param image: Input image.
    :type image: ImageTyping
    :param model: Model name for DBNET or DBNET++.
    :type model: str
    :param threshold: Confidence threshold for text detection.
    :type threshold: float
    :param max_area_size: Max area size when doing inference. Default is ``640``, which means if
        the image's area is over 640x640, it will be resized. When assigned to ``None``,
        it means do not resize in any case.
    :type max_area_size: Optional[int]
    :return: List of detected text bounding boxes, labels, and scores.
    :rtype: List[Tuple[Tuple[int, int, int, int], str, float]]

    .. warning::
        This function is deprecated, and it will be removed from imgutils in the future.
        Please migrate to :func:`imgutils.ocr.detect_text_with_ocr` as soon as possible.
    """
    image = load_image(image, mode='RGB')
    if max_area_size is not None and image.width * image.height >= max_area_size ** 2:
        r = ((image.width * image.height) / (max_area_size ** 2)) ** 0.5
        new_width, new_height = int(image.width / r), int(image.height / r)
        image = image.resize((new_width, new_height))
    else:
        r = 1.0

    bboxes = []
    for (x0, y0, x1, y1), score in _get_bounding_box_of_text(image, model, threshold):
        x0, y0, x1, y1 = int(x0 * r), int(y0 * r), int(x1 * r), int(y1 * r)
        bboxes.append(((x0, y0, x1, y1), 'text', score))

    bboxes = sorted(bboxes, key=lambda x: x[2], reverse=True)
    return bboxes
